from .onclickInterface import OnclickInterface
from optimeed.core import printIfShown, SHOW_WARNING
from optimeed.visualize.mainWindow import MainWindow
import traceback
from abc import ABCMeta, abstractmethod


class RepresentDeviceInterface(metaclass=ABCMeta):
    @abstractmethod
    def get_widget(self, theDevice):
        """Get Qt widget that represents the device

        :param theDevice: the Device to be represented
        :return: Qt widget
        """
        pass


class Onclick_representDevice(OnclickInterface):
    """On click: show informations about the points (loop through attributes)"""

    class DataInformationVisuals:
        def __init__(self):
            self.listOfVisuals = {}
            self.theTraces = {}
            self.indicesPoints = {}
            self.index = 0

        def delete_visual(self, theVisual):
            goodkey = 0
            for key in self.listOfVisuals:
                if self.listOfVisuals[key] == theVisual:
                    goodkey = key
            self.theTraces[goodkey].reset_brush(self.indicesPoints[goodkey])

        def add_visual(self, theVisual, theTrace, indexPoint):
            theTrace.set_brush(indexPoint, (250, 250, 0))

            index = self.get_new_index()
            self.listOfVisuals[index] = theVisual
            self.indicesPoints[index] = indexPoint
            self.theTraces[index] = theTrace

            theVisual.move(10, 0)
            theVisual.setWindowTitle("Visual information of point " + str(index))
            # Put all previously opened window on top
            for key in self.listOfVisuals:
                self.listOfVisuals[key].raise_()

        def get_new_index(self):
            self.index += 1
            return self.index

        def curr_index(self):
            return self.index

    def __init__(self, theLinkDataGraph, visuals):
        """

        :param theLinkDataGraph: :class:`~optimeed.core.linkDataGraph.LinkDataGraph`
        :param visuals: List of classes following class:`RepresentDeviceInterface`
        """
        self.theLinkDataGraph = theLinkDataGraph
        self.dataInformationVisuals = self.DataInformationVisuals()
        self.visuals = visuals

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        """Action to perform when a point in the graph has been clicked:
        Creates new window displaying the device and its informations
        """
        try:
            def actionOnWindowClosed(theVisual, _):
                """Action to perform when a window has been closed:
                Remove the window from the dataInformationVisuals
                """
                theVisual.close()
                self.dataInformationVisuals.delete_visual(theVisual)

            theTrace = theGraphVisual.get_graph(index_graph).get_trace(index_trace)
            for index_point in indices_points:
                theDevice = self.theLinkDataGraph.get_clicked_item(index_graph, index_trace, index_point)

                theWidgetList = list()
                for visual in self.visuals:
                    theWidgetList.append(visual.get_widget(theDevice))

                visual_temp = MainWindow(theWidgetList, actionOnWindowClosed=actionOnWindowClosed)
                self.dataInformationVisuals.add_visual(visual_temp, theTrace, index_point)
                visual_temp.run(False)
        except KeyboardInterrupt:
            raise
        except Exception:
            printIfShown("Following error occurred in visualisation :" + traceback.format_exc(), SHOW_WARNING)

    def get_name(self):
        return "Represent Device"



SHOW_WARNING = 0
SHOW_INFO = 1
SHOW_ERROR = 2
SHOW_DEBUG = 3
SHOW_LOGS = 4
SHOW_CURRENT = [SHOW_INFO, SHOW_WARNING, SHOW_ERROR, SHOW_DEBUG, SHOW_LOGS]
# SHOW_CURRENT = []


def setCurrentShow(show_types):
    """Change text type to be displayed by PrintIfShown"""
    global SHOW_CURRENT
    SHOW_CURRENT.clear()
    for t in show_types:
        if 0 <= t < 5 and isinstance(t, int):
            SHOW_CURRENT.append(t)


def getCurrentShow():
    """Get text type to be displayed by PrintIfShown"""
    global SHOW_CURRENT
    return list(SHOW_CURRENT)


def disableLogs():
    """Disable all logs"""
    global SHOW_CURRENT
    try:
        SHOW_CURRENT.remove(SHOW_LOGS)
    except ValueError:
        pass


def enableLogs():
    """Show all logs"""
    global SHOW_CURRENT
    if SHOW_LOGS not in SHOW_CURRENT:
        SHOW_CURRENT.append(SHOW_LOGS)



from optimeed.visualize.widgets import Widget_openGL, Widget_lineDrawer, Widget_image, Widget_text_scrollable
from optimeed.core import rgetattr, str_all_attr
from .onclick_representDevice import RepresentDeviceInterface

class Represent_lines(RepresentDeviceInterface):
    def __init__(self, attribute_lines):
        """Uses Widget_lineDrawer to display the lines contained in attribute "attribute_lines"

        :param attribute_lines: name of attributes where the lines are stored
        """
        self.attribute_lines = attribute_lines

    def get_widget(self, theNewDevice):
        widgetLineDrawer = Widget_lineDrawer()
        widgetLineDrawer.set_lines(rgetattr(theNewDevice, self.attribute_lines))
        return widgetLineDrawer


class Represent_txt_function(RepresentDeviceInterface):
    def __init__(self, is_light=True, convertToHtml=True):
        """Display some text from the device

        :param is_light: Use light background
        :param convertToHtml: Convert text to html (for color support)
        """
        self.is_light = is_light
        self.convertToHtml = convertToHtml

    def getTxt(self, theNewDevice):
        return "Ici va le texte pour l'objet --> " + str(theNewDevice)

    def get_widget(self, theNewDevice):
        return Widget_text_scrollable(self.getTxt(theNewDevice), is_light=self.is_light, convertToHtml=self.convertToHtml)


class Represent_brut_attributes(RepresentDeviceInterface):
    def __init__(self, is_light=True, convertToHtml=True, recursion_level=5):
        """Loop through the attributes and display them as plain text

        :param is_light: Use light background
        :param convertToHtml: Convert text to html (for color support)
        :param recursion_level: Level of nesting to display
        """
        self.is_light = is_light
        self.convertToHtml = convertToHtml
        self.recursion_level = recursion_level

    def get_widget(self, theNewDevice):
        return Widget_text_scrollable(str_all_attr(theNewDevice, self.recursion_level), is_light=self.is_light, convertToHtml=self.convertToHtml)


class Represent_opengl(RepresentDeviceInterface):
    def __init__(self, DeviceDrawer):
        """
        Use openGL to represent the device

        :param DeviceDrawer: OpenGL drawer, see
        """

        self.DeviceDrawer = DeviceDrawer

    def get_widget(self, theNewDevice):
        theOpenGLWidget = Widget_openGL()
        theOpenGLWidget.set_deviceDrawer(self.DeviceDrawer)
        theOpenGLWidget.set_deviceToDraw(theNewDevice)
        return theOpenGLWidget


class Represent_image(RepresentDeviceInterface):
    def __init__(self, get_base_64_from_device):
        """
        Display an image (png) representation of the motor

        :param get_base_64_from_device: method taking a device as input and returns a base64 string
        """
        self.get_base_64_from_device = get_base_64_from_device

    def get_widget(self, theNewDevice):
        return Widget_image(self.get_base_64_from_device(theNewDevice))




"""
Not from us
"""


def IsConvex(a, b, c):
    # only convex if traversing anti-clockwise!
    crossp = (b[0] - a[0]) * (c[1] - a[1]) - (b[1] - a[1]) * (c[0] - a[0])
    if crossp >= 0:
        return True
    return False


def InTriangle(a, b, c, p):
    L = [0, 0, 0]
    eps = 0.0000001
    # calculate barycentric coefficients for point p
    # eps is needed as error correction since for very small distances denom->0
    L[0] = ((b[1] - c[1]) * (p[0] - c[0]) + (c[0] - b[0]) * (p[1] - c[1])) / (((b[1] - c[1]) * (a[0] - c[0]) + (c[0] - b[0]) * (a[1] - c[1])) + eps)
    L[1] = ((c[1] - a[1]) * (p[0] - c[0]) + (a[0] - c[0]) * (p[1] - c[1])) / (((b[1] - c[1]) * (a[0] - c[0]) + (c[0] - b[0]) * (a[1] - c[1])) + eps)
    L[2] = 1 - L[0] - L[1]
    # check if p lies in triangle (a, b, c)
    for x in L:
        if x >= 1 or x <= 0:
            return False
    return True


def IsClockwise(poly):
    # initialize sum with last element
    sum = (poly[0][0] - poly[len(poly) - 1][0]) * (poly[0][1] + poly[len(poly) - 1][1])
    # iterate over all other elements (0 to n-1)
    for i in range(len(poly) - 1):
        sum += (poly[i + 1][0] - poly[i][0]) * (poly[i + 1][1] + poly[i][1])
    if sum > 0:
        return True
    return False


def GetEar(poly):
    size = len(poly)
    if size < 3:
        return []
    if size == 3:
        tri = (poly[0], poly[1], poly[2])
        del poly[:]
        return tri
    for i in range(size):
        tritest = False
        p1 = poly[(i - 1) % size]
        p2 = poly[i % size]
        p3 = poly[(i + 1) % size]
        if IsConvex(p1, p2, p3):
            for x in poly:
                if not (x in (p1, p2, p3)) and InTriangle(p1, p2, p3, x):
                    tritest = True
            if not tritest:
                del poly[i % size]
                return p1, p2, p3
    print('GetEar(): no ear found')
    return []


def reformatXYtoList(xList, yList):
    polyList = [None]*len(xList)
    for i in range(len(xList)):
        polyList[i] = [xList[i], yList[i]]
    return polyList


def meshPolygon(xList, yList):
    pListIn = reformatXYtoList(xList, yList)
    triList = []
    plist = pListIn[::-1] if IsClockwise(pListIn) else pListIn[:]
    while len(plist) >= 3:
        a = GetEar(plist)
        if not a:
            break
        triList.append(a)
    return triList






from abc import ABCMeta, abstractmethod


class InterfaceMathsToPhysics(metaclass=ABCMeta):
    """Interface to transform output from the optimizer to meaningful variables of the device"""

    @abstractmethod
    def fromMathsToPhys(self, xVector, theDevice, opti_variables):
        """
        Transforms an input vector coming from the optimization (e.g. [0.23, 4, False]) to "meaningful" variable (ex: length, number of poles, flag).

        :param xVector: List of optimization variables from the optimizer
        :param theDevice: :class:`~optimeed.InterfaceDevice.InterfaceDevice`
        :param opti_variables: list of :class:`~optimeed.optimize.OptimizationVariable.OptimizationVariable`
        """
        pass

    @abstractmethod
    def fromPhysToMaths(self, theDevice, opti_variables):
        """
        Extracts a mathematical vector from meaningful variable of the Device

        :param theDevice: :class:`~optimeed.InterfaceDevice.InterfaceDevice`
        :param opti_variables: list of :class:`~optimeed.optimize.OptimizationVariable.OptimizationVariable`
        :return: List of optimization variables
        """
        pass



from abc import abstractmethod, ABCMeta


class InterfaceConvergence(metaclass=ABCMeta):
    """Simple interface to visually get the convergence of any optimization problem"""

    @abstractmethod
    def get_graphs(self, *args, **kwargs):
        """Return :class:`~optimeed.core.graphs.Graphs`"""
        pass



try:
    from scipy.interpolate import griddata
    griddata_found = True
except ImportError:
    griddata_found = False
import numpy as np


class Plot3D_Generic:
    def __init__(self, x_label='', y_label='', z_label='', legend='',
                 x_lim=None, y_lim=None,z_lim=None):
        self.x_label = x_label
        self.y_label = y_label
        self.z_label = z_label
        self.x_lim = x_lim
        self.y_lim = y_lim
        self.z_lim = z_lim
        self.legend = legend

    def get_lim(self, axis):
        return getattr(self, "{}_lim".format(axis))

    def get_label(self, axis):
        return getattr(self, "{}_label".format(axis))

    def get_legend(self):
        return self.legend


class GridPlot_Generic(Plot3D_Generic):
    def __init__(self, X, Y, Z, **kwargs):
        super().__init__(**kwargs)
        self.X = X
        self.Y = Y
        self.Z = Z

    def get_plot_data(self):
        return self.X, self.Y, self.Z


class ContourPlot(GridPlot_Generic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.numberOfContours = kwargs.get("n", 10)
        self.levels = kwargs.get("levels", list())

    def get_levels(self):
        return self.levels

    def get_number_of_contours(self):
        return self.numberOfContours


class FilledContourPlot(ContourPlot):
    pass


class SurfPlot(GridPlot_Generic):
    pass


class MeshPlot(GridPlot_Generic):
    pass


class ScatterPlot3(Plot3D_Generic):
    def __init__(self, x, y, z, **kwargs):
        self.color = kwargs.pop("color", None)
        super().__init__(**kwargs)
        self.x = x
        self.y = y
        self.z = z

    def get_plot_data(self):
        return self.x, self.y, self.z

    def get_color(self):
        return self.color


def convert_to_gridplot(x, y, z, x_interval=None, y_interval=None, n_x=20, n_y=20):
    """
    Convert set of points x, y, z to a grid

    :param x:
    :param y:
    :param z:
    :param x_interval: [Min, max] of the grid. If none, use min and max values
    :param y_interval: [Min, max] of the grid.  If none, use min and max values
    :param n_x: number of points in x direction
    :param n_y: number of points in y direction
    :return: X, Y, Z as grid
    """
    if griddata_found:

        if x_interval is None:
            x_interval = [min(x), max(x)]
        if y_interval is None:
            y_interval = [min(y), max(y)]
        xi = np.linspace(*x_interval, n_x)
        yi = np.linspace(*y_interval, n_y)
        zi = griddata((x, y), z, (xi[None, :], yi[:, None]), method='cubic')  # 2D array
        return xi, yi, zi
    else:
        raise ImportError("Scipy is needed for that")



import numpy as np
from multiprocessing import Pool
import time


def _is_feasible(theList):
    return np.all(np.array(theList) <= 0)


def _format_fx_fs(objectives_pop, constraints_pop):
    fx = np.zeros(len(objectives_pop))
    fs = np.zeros(len(constraints_pop), dtype=bool)
    for i in range(len(fx)):
        fx[i] = objectives_pop[i][0]
        fs[i] = _is_feasible(constraints_pop[i])
    return fx, fs


class MyMapEvaluator:
    def __init__(self, evaluation_function, callback_on_evaluation):
        self.callback_on_evaluation = callback_on_evaluation
        self.evaluation_function = evaluation_function

    def evaluate_all(self, x):
        objectives_pop = [None]*len(x)
        constraints_pop = [None]*len(x)

        for k, xi in enumerate(x):
            returned_value = self.evaluation_function(xi)
            self.callback_on_evaluation(returned_value)

            objectives_pop[k] = returned_value["objectives"]
            constraints_pop[k] = returned_value["constraints"]
        return objectives_pop, constraints_pop


class MyMultiprocessEvaluator:
    def __init__(self, evaluation_function, callback_on_evaluation, numberOfCores):
        self.callback_on_evaluation = callback_on_evaluation
        self.evaluation_function = evaluation_function
        self.pool = Pool(numberOfCores)

    def evaluate_all(self, x):
        objectives_pop = [None]*len(x)
        constraints_pop = [None]*len(x)
        for k, returned_value in enumerate(self.pool.imap_unordered(self.evaluation_function, x)):
            self.callback_on_evaluation(returned_value)
            objectives_pop[k] = returned_value["objectives"]
            constraints_pop[k] = returned_value["constraints"]
        return objectives_pop, constraints_pop


def pso(lb, ub, initialVectorGuess, theEvaluator, maxtime, callback_generation=lambda objectives, constraints: None, swarmsize=100, omega=0.5, phip=0.5, phig=0.5):
    """
    Perform a particle swarm optimization (PSO)

    Parameters
    ==========
    lb: list
        Lower bounds of each parameter
    ub: list
        upper bounds of each parameter
    initialVectorGuess: list
        initial vector guess for the solution (to be included inside population)
    theEvaluator : object define before
    maxtime : float
        The maximum time (in s) before stopping the algorithm
    callback_generation: function lambda (bjectives (as list), constraints (as list)) per step
        Useful to log convergence
    swarmsize : int
        The number of particles in the swarm (Default: 100)
    omega : scalar
        Particle velocity scaling factor (Default: 0.5)
    phip : scalar
        Scaling factor to search away from the particle's best known position
        (Default: 0.5)
    phig : scalar
        Scaling factor to search away from the swarm's best known position
        (Default: 0.5)

    Returns
    =======
    g : array
        The swarm's best known position (optimal design)
    f : scalar
        The objective value at ``g``
    """
    lb = np.array(lb)
    ub = np.array(ub)

    vhigh = np.abs(ub - lb)
    vlow = -vhigh

    # Initialize the particle swarm ############################################
    S = swarmsize
    D = len(lb)  # the number of dimensions each particle has
    x = np.random.rand(S, D)  # particle positions
    # v = np.zeros_like(x)  # particle velocities
    p = np.zeros_like(x)  # best particle positions
    # fx = np.zeros(S)  # current particle function values
    # fs = np.zeros(S, dtype=bool)  # feasibility of each particle
    fp = np.ones(S)*np.inf  # best particle function values
    # g = []  # best swarm position
    fg = np.inf  # best swarm position starting value
    fsg = True

    # Initialize the particle's position
    x = lb + x*(ub - lb)

    x[0, :] = initialVectorGuess[:]  # inserted initial guess in population

    # Calculate objective and constraints for each particle
    obj, cons = theEvaluator.evaluate_all(x)
    fx, fs = _format_fx_fs(obj, cons)

    # Store particle's best position (if constraints are satisfied)
    i_update = np.logical_and((fx < fp), fs)
    p[i_update, :] = x[i_update, :].copy()
    fp[i_update] = fx[i_update]

    # Update swarm's best position
    i_min = np.argmin(fp)
    if fp[i_min] < fg:
        fg = fp[i_min]
        g = p[i_min, :].copy()
    else:
        # At the start, there may not be any feasible starting point, so just
        # give it a temporary "best" point since it's likely to change
        g = x[0, :].copy()

    # Initialize the particle's velocity
    v = vlow + np.random.rand(S, D)*(vhigh - vlow)

    # Iterate until termination criterion met ##################################

    start = time.time()
    while (time.time() - start) <= maxtime:
        rp = np.random.uniform(size=(S, D))
        rg = np.random.uniform(size=(S, D))

        # Update the particles velocities
        v = omega*v + phip*rp*(p - x) + phig*rg*(g - x)
        # Update the particles' positions
        x = x + v
        # Correct for bound violations
        maskl = x < lb
        masku = x > ub
        x = x*(~np.logical_or(maskl, masku)) + lb*maskl + ub*masku

        # Update objectives and constraints
        obj, cons = theEvaluator.evaluate_all(x)
        fx, fs = _format_fx_fs(obj, cons)

        callback_generation(obj, cons)
        # Store particle's best position (if constraints are satisfied)
        i_update = np.logical_and((fx < fp), fs)
        p[i_update, :] = x[i_update, :].copy()
        fp[i_update] = fx[i_update]

        # Compare swarm's best position with global best position
        i_min = np.argmin(fp)
        if fp[i_min] < fg:
            p_min = p[i_min, :].copy()
            g = p_min.copy()
            fg = fp[i_min]
            fsg = fs[i_min]

    return g, fg, fsg



from .onclick_representDevice import Onclick_representDevice, RepresentDeviceInterface
from .onclick_animate import Onclick_animate
from .onclick_changeSymbol import Onclick_changeSymbol
from .onclick_copySomething import Onclick_copySomething
from .onclick_delete import Onclick_delete
from .onclick_exportCollection import Onclick_exportCollection
from .onclick_exportToTxt import Onclick_exportToTxt
from .onclick_exportTrace import Onclick_exportTrace
from .onclick_extractPareto import Onclick_extractPareto
from .onclick_measure import Onclick_measure
from .onclick_removeTrace import Onclick_removeTrace
from .onclick_representDevice import Onclick_representDevice
from .onclick_tojson import Onclick_tojson
from .onclickInterface import OnclickInterface
from .representDevice_examples import Represent_opengl, Represent_image, Represent_lines, \
    Represent_brut_attributes, Represent_txt_function
from .animation_examples import Animate_lines, Animate_openGL, Animate_lines_and_text, Animate_openGL_and_text



from PyQt5 import QtCore, QtWidgets, QtSvg, QtGui
from abc import abstractmethod
import subprocess
import os
import math


class _AnimationTrace:
    """Contains all the element to animate for a trace"""
    class AnimationElement:
        def __init__(self, elements):
            self.elements = elements

        def get(self):
            return self.elements

    def __init__(self, elements_list, theTrace):
        self.curr_index = 0
        self.indices_to_show = []

        self.element_animations = []
        for elements in elements_list:
            self.element_animations.append(self.AnimationElement(elements))

        self.theTrace = theTrace
        self.nbr_elements = len(self.element_animations[0].get())

    def get_element_animations(self, itemNumber, index_in_show):
        """
        Get the element to show
        :param itemNumber: item number (0 if only one think to draw)
        :param index_in_show: index in the list
        :return: The element to draw
        """
        return self.element_animations[itemNumber].get()[self.map_index(index_in_show)]

    def show_all(self):
        self.delete_all()
        self.indices_to_show = list(range(self.nbr_elements))
        self.curr_index = 0

    def delete_all(self):
        self.theTrace.reset_all_brushes()
        self.indices_to_show = []
        self.curr_index = 0

    def get_indices_to_show(self):
        return self.indices_to_show

    def add_element(self, indexPoint):
        if indexPoint not in self.get_indices_to_show():
            self.add_index_to_show(indexPoint)
            self.theTrace.set_brush(indexPoint, (250, 250, 0))
        else:
            self._remove_index_from_show(indexPoint)

    def add_index_to_show(self, index):
        self.indices_to_show.append(index)

    def _remove_index_from_show(self, index):
        self.theTrace.reset_brush(index)
        self.get_indices_to_show().remove(index)

    def set_curr_brush(self, index_in_show):
        self.theTrace.set_brush(self.map_index(index_in_show), (0, 0, 250))

    def set_idle_brush(self, index_in_show):
        self.theTrace.set_brush(self.map_index(index_in_show), self.theTrace.get_base_symbol_brush().color().lighter(150))

    def get_number_of_elements(self):
        return len(self.get_indices_to_show())

    def map_index(self, index_in_show):
        try:
            return self.get_indices_to_show()[index_in_show]
        except IndexError:
            return 0

    def get_base_pen(self):
        return self.theTrace.get_base_pen()


class AnimationGUI(QtWidgets.QMainWindow):
    """Spawns a gui that includes button to create animations nicely when paired with :class:`~optimeed.visualize.gui.widgets.widget_graphs_visual`"""
    SlIDER_MAXIMUM_VALUE = 500
    SLIDER_MINIMUM_VALUE = 1

    def __init__(self, id=0, window_title='Animation'):
        """

        :param id: id of the window (int)
        :param window_title: Title of the window (str)
        """

        self.refresh_time = 0.1
        self.isRunning = False
        self.data_animation_traces = dict()
        self.id = id

        # Create PyQt5 window
        super().__init__()
        self.mainbox = QtWidgets.QWidget()
        self.setWindowTitle(window_title)
        self.setCentralWidget(self.mainbox)
        self.mainbox.setLayout(QtWidgets.QVBoxLayout())
        self.mainbox.layout().setContentsMargins(0, 0, 0, 0)

        self.horizontal_layout_visu = QtWidgets.QHBoxLayout()
        self.mainbox.layout().addLayout(self.horizontal_layout_visu)

        # Add button panels
        horizontal_layout = QtWidgets.QHBoxLayout()

        # First: Show all
        button = QtWidgets.QPushButton('Show All', self)
        button.clicked.connect(self.show_all)
        horizontal_layout.addWidget(button)

        # Second: Delete All
        button = QtWidgets.QPushButton('Delete All', self)
        button.clicked.connect(self.delete_all)
        horizontal_layout.addWidget(button)

        # Third: Pause
        self.isPlaying = True
        self.pause_play_button = QtWidgets.QPushButton('Pause', self)
        self.pause_play_button.clicked.connect(self.pause_play)
        horizontal_layout.addWidget(self.pause_play_button)

        # Fourth: Speed
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.setMinimum(self.SLIDER_MINIMUM_VALUE)
        self.slider.setMaximum(self.SlIDER_MAXIMUM_VALUE)
        self.slider.valueChanged.connect(self.slider_handler)
        self.mainbox.layout().addWidget(self.slider)

        # Put them in layout
        self.mainbox.layout().addLayout(horizontal_layout)

        # Add export option
        exportButton = QtWidgets.QPushButton('Export', self)
        exportButton.clicked.connect(self.export_picture)
        self.mainbox.layout().addWidget(exportButton)

    def add_trace(self, trace_id, element_list, theTrace):
        """
        Add a trace to the animation.

        :param trace_id: id of the trace
        :param element_list: List of elements to save: [[OpenGL_item1, text_item1], [OpenGL_item2, text_item2], ... [OpenGL_itemN, text_itemN]]
        :param theTrace: :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.TraceVisual.TraceVisual`
        :return:
        """
        self.data_animation_traces[trace_id] = _AnimationTrace(self.get_interesting_elements(element_list), theTrace)

    @staticmethod
    def get_interesting_elements(element_list):
        """
        Function called upon new trace creation. From a list, takes the interesting elements for animation
        :param element_list:
        :return: new_element_list
        """
        return [element_list]

    def add_elementToTrace(self, trace_id, indexPoint):
        self.data_animation_traces[trace_id].add_element(indexPoint)

    def delete_point(self, trace_id, thePoint):
        self.data_animation_traces[trace_id].delete_point(thePoint)

    def reset_all(self):
        for trace_id in self.data_animation_traces:
            self.data_animation_traces[trace_id].delete_all()

    def delete_all(self):
        self.isRunning = False
        self.reset_all()
        for key in self.data_animation_traces:
            self.delete_key_widgets(key)

        self.data_animation_traces = dict()

    def pause_play(self):
        if self.isRunning:
            self.pause_play_button.setText("Play")
            self.isRunning = False
        else:
            self.pause_play_button.setText("Pause")
            self.run()
            self.isRunning = True

    def show_all(self):
        buff_state_running = self.isRunning
        self.reset_all()
        for key in self.data_animation_traces:
            self.data_animation_traces[key].show_all()
        self.isRunning = buff_state_running
        self.run()

    def next_frame(self):
        if self.isRunning:
            try:
                for key in self.data_animation_traces:
                    the_data_animation = self.data_animation_traces[key]
                    the_data_animation.set_idle_brush(the_data_animation.curr_index)

                    the_data_animation.curr_index += 1
                    the_data_animation.curr_index = the_data_animation.curr_index % the_data_animation.get_number_of_elements()

                    the_data_animation.set_curr_brush(the_data_animation.curr_index)
                    self.update_widget_w_animation(key, the_data_animation.curr_index, the_data_animation)
            except (AttributeError, ZeroDivisionError):
                pass

            QtCore.QTimer().singleShot(int(self.refresh_time*1000), self.next_frame)

    def slider_handler(self):
        if self.isRunning:
            self.set_refreshTime()
        else:
            self.frame_selector()

    def frame_selector(self):
        for key in self.data_animation_traces:
            data_animation_trace = self.data_animation_traces[key]
            slider_value = self.slider.value()
            a = (data_animation_trace.get_number_of_elements()-1)/(self.SlIDER_MAXIMUM_VALUE - self.SLIDER_MINIMUM_VALUE)
            b = -a*self.SLIDER_MINIMUM_VALUE
            selected_index = min(math.floor(a*slider_value+b), data_animation_trace.get_number_of_elements()-1)

            self.data_animation_traces[key].set_idle_brush(data_animation_trace.curr_index)
            data_animation_trace.curr_index = selected_index
            data_animation_trace.set_curr_brush(data_animation_trace.curr_index)
            self.update_widget_w_animation(key, selected_index, data_animation_trace)

    def set_refreshTime(self):
        self.refresh_time = self.slider.value()/1000

    def is_empty(self):
        is_empty = True
        for key in self.data_animation_traces:
            if len(self.data_animation_traces[key].get_indices_to_show()):
                is_empty = False
        return is_empty

    def run(self):
        if (not self.isRunning) and (not self.is_empty()):
            self.show()
            self.isRunning = True
            self.next_frame()

    def closeEvent(self, _):
        self.isRunning = False
        self.delete_all()

    def contains_trace(self, trace_id):
        return trace_id in self.data_animation_traces

    def export_picture(self):
        dlg = QtWidgets.QFileDialog.getSaveFileName()[0]
        if dlg:
            root, ext = os.path.splitext(dlg)
            filename_png = root + '.png'
            filename_svg = root + '.svg'

            # Create exporter
            generator = QtSvg.QSvgGenerator()
            generator.setFileName(filename_svg)
            painter = QtGui.QPainter()
            painter.begin(generator)
            self.export_widget(painter)
            painter.end()

            with open(filename_svg, 'r') as f:
                lines = f.read()
                lines = lines.replace('vector-effect="non-scaling-stroke"', 'stroke-width="2"')
            with open(filename_svg, 'w') as f:
                f.write(lines)
            converter_cmd_line = 'inkscape -z ' + filename_svg + ' -D -e ' + filename_png + ' -d 400'
            subprocess.Popen(converter_cmd_line, shell=True)

    @abstractmethod
    def export_widget(self, painter):
        """
        Render scene with a painter

        :param painter: PyQt painter
        """
        pass

    @abstractmethod
    def update_widget_w_animation(self, key, index, the_data_animation):
        """
        What to do when a new element has to be animated. Example: self.theOpenGLWidget.set_deviceToDraw(the_data_animation.get_element_animations(0, index))

        :param key: key of the trace that has to be animated
        :param index: index that has to be animated
        :param the_data_animation: :class:`~DataAnimationTrace` that has to be animated
        """
        pass

    @abstractmethod
    def delete_key_widgets(self, key):
        """
        What to do when a key has to be deleted

        :param key: key of the trace that has to be deleted
        """
        pass



import signal
import sys
from PyQt5 import QtWidgets

if QtWidgets.QApplication.instance() is None:
    app = QtWidgets.QApplication(sys.argv)  # must initialize only once


def start_qt_mainloop():
    """Starts qt mainloop, which is necessary for qt to handle events"""

    def handler_quit(sign_number, _):
        app.quit()
        sys.exit(sign_number)

    try:
        signal.signal(signal.SIGINT, handler_quit)
    except AttributeError:
        pass
    try:
        signal.signal(signal.SIGTSTP, handler_quit)
    except AttributeError:
        pass
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app.exec()


def stop_qt_mainloop():
    """Stops qt mainloop and resumes to program"""
    app.quit()


def process_qt_events():
    """Process current qt events"""
    app.processEvents()



from .onclickInterface import OnclickInterface
from optimeed.core.tools import get_2D_pareto
from optimeed.core.collection import ListDataStruct


class Onclick_extractPareto(OnclickInterface):
    """On click: extract the pareto from the cloud of points"""
    def __init__(self, theDataLink, max_x=False, max_y=False):
        """

        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        :param max_x: x axis is to maximize or not (bool)
        :param max_y: y axis is to maximize or not (bool)
        """
        self.theDataLink = theDataLink
        self.max_x = max_x
        self.max_y = max_y

    def graph_clicked(self, the_graph_visual, index_graph, index_trace, _):
        # Retrieve collection plotted -> getShadow = False
        theCollection_plotted = self.theDataLink.get_collection_from_graph(index_graph, index_trace, getShadow=False)

        # Recover x-y points
        x_list, y_list = self.theDataLink.get_x_y_to_plot(theCollection_plotted, self.theDataLink.get_howToPlotGraph(index_graph))

        # Get pareto from them
        xx, yy, indices = get_2D_pareto(x_list, y_list, max_X=self.max_x, max_Y=self.max_y)

        # Create new collection. We extract the "X-Y" (non-shadow) points from original collection
        newCollection_plotted = ListDataStruct()
        newCollection_plotted.set_data([theCollection_plotted.get_data_at_index(index) for index in indices])

        # Add it to the data link
        id_newCollection_plotted = self.theDataLink.add_collection(newCollection_plotted, {"is_scattered": False, "sort_output": True, "legend": 'Pareto extracted'})

        # Now we also need to extract the underlying motors.
        theCollection_motor = self.theDataLink.get_collection_from_graph(index_graph, index_trace, getShadow=True)
        subset_collection_motor = theCollection_motor.extract_collection_from_indices(indices)

        # We add it to the datalink as shadow
        self.theDataLink.set_shadow_collection(id_newCollection_plotted, subset_collection_motor)

    def get_name(self):
        return "Extract pareto"



from PyQt5.QtGui import QIcon
from PyQt5.QtDesigner import QPyDesignerCustomWidgetPlugin
from .widget_tableWithSearch import Widget_tableWithSearch as TheWidget


class Plugin_tableWithSearch(QPyDesignerCustomWidgetPlugin):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.initialized = False

    def initialize(self, core):
        if self.initialized:
            return

        self.initialized = True

    def isInitialized(self):
        return self.initialized

    def createWidget(self, parent):
        return TheWidget(parent)

    def name(self):
        return "Widget_table_with_search"

    def group(self):
        return "ICS Custom Widgets"

    def icon(self):
        return QIcon()

    def toolTip(self):
        return ""

    def whatsThis(self):
        return ""

    def isContainer(self):
        return False

    # Returns an XML description of a custom widget instance that describes
    # default values for its properties. Each custom widget created by this
    # plugin will be configured using this description.
    # def domXml(self):
    #     return '<widget class="PyAnalogClock" name="analogClock">\n' \
    #            ' <property name="toolTip">\n' \
    #            '  <string>The current time</string>\n' \
    #            ' </property>\n' \
    #            ' <property name="whatsThis">\n' \
    #            '  <string>The analog clock widget displays the current ' \
    #            'time.</string>\n' \
    #            ' </property>\n' \
    #            '</widget>\n'

    def includeFile(self):
        return "optimeed.visualize.widgets.widget_tableWithSearch"






from optimeed.core.tools import text_format, rsetattr, rgetattr, printIfShown, SHOW_WARNING


class OptimizationVariable:
    """Contains information about the optimization of a variable"""
    attributeName: str

    def __init__(self, attributeName):
        """Instantiates the opti variable

        :param attributeName: name of the attribute to optimize
        """
        self.attributeName = attributeName

    def get_attribute_name(self):
        """Return the attribute to set"""
        return self.attributeName

    def add_prefix_attribute_name(self, thePrefix):
        """Used for nested object, lower the name by prefix. Example: R_ext becomes (thePrefix).R_ext"""
        self.attributeName = "{}.{}".format(thePrefix, self.get_attribute_name())

    def get_PhysToMaths(self, deviceIn):
        """Convert the initial value of the variable contained in the device to optimization variable value

        :param deviceIn: :class:`~optimeed.InterfaceDevice.InterfaceDevice`
        :return: value of the corresponding optimization variable
        """
        return rgetattr(deviceIn, self.get_attribute_name())

    def do_MathsToPhys(self, variableValue, deviceIn):
        """Apply the value to the device"""
        rsetattr(deviceIn, self.get_attribute_name(), variableValue)

    def __str__(self):
        return "Name: {:^25s} ".format(text_format.GREEN + self.attributeName + text_format.END)


class Real_OptimizationVariable(OptimizationVariable):
    """Real (continuous) optimization variable. Most used type"""
    val_min: float
    val_max: float

    def __init__(self, attributeName, val_min, val_max):
        """
        Instantiates the opti variable

        :param attributeName: name of the attribute to optimize
        :param val_min: minimal value of the optimization variable
        :param val_max: maximal value of the optimization variable
        """
        super().__init__(attributeName)
        self.val_min = val_min
        self.val_max = val_max

    def get_min_value(self):
        return self.val_min

    def get_max_value(self):
        return self.val_max

    def get_PhysToMaths(self, deviceIn):
        value = super().get_PhysToMaths(deviceIn)
        if not self.get_min_value() <= value <= self.get_max_value():
            printIfShown("Initial individual breaks design space boundaries: {}."
                         " The problem has been solved by using middle interval value".format(self.get_attribute_name()), SHOW_WARNING)
            return (self.get_min_value() + self.get_max_value())/2
        return value

    def do_MathsToPhys(self, value, deviceIn):
        super().do_MathsToPhys(float(value), deviceIn)

    def __str__(self):
        return '{} Real variable | Min Value: {:>5s} \t Max Value: {:<15s}  \n'.format(str(super().__str__()), str(self.get_min_value()), str(self.get_max_value()))


class Binary_OptimizationVariable(OptimizationVariable):
    """Boolean (True/False) optimization variable."""

    def get_PhysToMaths(self, deviceIn):
        return bool(super().get_PhysToMaths(deviceIn))

    def do_MathsToPhys(self, value, deviceIn):
        super().do_MathsToPhys(bool(value), deviceIn)

    def __str__(self):
        return '{} Binary variable |  \n'.format(str(super().__str__()))


class Integer_OptimizationVariable(OptimizationVariable):
    """Integer variable, in [min_value, max_value]"""

    val_min: int
    val_max: int

    def __init__(self, attributeName, val_min, val_max):
        """
        Instantiates the opti variable

        :param attributeName: name of the attribute to optimize
        :param val_min: minimal value of the optimization variable
        :param val_max: maximal value of the optimization variable
        """
        super().__init__(attributeName)
        self.val_min = val_min
        self.val_max = val_max

    def get_min_value(self):
        return self.val_min

    def get_max_value(self):
        return self.val_max

    def get_PhysToMaths(self, deviceIn):
        value = super().get_PhysToMaths(deviceIn)
        if not self.get_min_value() <= value <= self.get_max_value():
            printIfShown("Initial individual breaks design space boundaries: {}."
                         " The problem has been solved by using middle interval value".format(self.get_attribute_name()), SHOW_WARNING)
            return int((self.get_min_value() + self.get_max_value()) / 2)
        return value

    def do_MathsToPhys(self, value, deviceIn):
        super().do_MathsToPhys(int(value), deviceIn)

    def __str__(self):
        return '{} Integer variable | Min Value: {:>5s} \t Max Value: {:<15s}  \n'.format(str(super().__str__()), str(self.get_min_value()), str(self.get_max_value()))



import ast
import inspect
from abc import ABCMeta
# from typing import Dict, List, Tuple
try:
    from typing import GenericMeta as _Generic
except ImportError:
    from typing import _GenericAlias as _Generic

import traceback
import numpy as np
from optimeed.core.tools import rgetattr, rsetattr, printIfShown, SHOW_ERROR, SHOW_WARNING
import sys
import os
import re


"""
Expected behaviour:
Object attributes values can either be of "known type" (=> type hinted) or "unknown type" (=> Saveable object).
- If type hinted, then save/load is pretty straightforward: 
    1. save module/class of parent
    2. Get typehinted attributes
    3. For each attribute, only save the value
    ----
    1. load module/class of parent
    2. Recover from annotations where the module/class of attributes are
    3. For each attribute, cast to this module/class
- If not, then save/load:
    1. Save module/class of parent
    2. For each attribute to save (get_additional_attributes_to_save):
    3. Save the value AND its module/class
    ----
    1. load module/class of parent
    2. Recover from parent the attributes that have been saved
    3. For each attribute, reconstitute the object using the added module/class information during save stage
"""

MODULE_TAG = '__module__'
CLASS_TAG = '__class__'
EXCLUDED_TAGS = [CLASS_TAG, MODULE_TAG]


def getExecPath():
    try:
        sFile = os.path.abspath(sys.modules['__main__'].__file__)
    except:
        sFile = sys.executable
    basename = os.path.basename(sFile)
    return os.path.splitext(basename)[0]


class SaveableObject(metaclass=ABCMeta):
    """Abstract class for dynamically type-hinted objects.
    This class is to solve the special case where the exact type of an attribute is not known before runtime, yet has to be saved."""

    def get_additional_attributes_to_save(self):
        """Return list of attributes corresponding to object, whose type cannot be determined statically (e.g. topology change)"""
        return list()

    def get_additional_attributes_to_save_list(self):
        """Same behavior as get_additional_attributes_to_save, but where the attributes contains list of unknown items"""
        return list()


def _isclass(theObject):
    """Extends the default isclass method with typing"""
    if isinstance(theObject, type):
        return True
    # Test if comes from typing class
    return isinstance(theObject, _Generic)


def get_type_class(typ):
    """Get the type of the class. used to compare objects from Typing."""
    try:
        # Python 3.5 / 3.6
        return typ.__extra__
    except AttributeError:
        try:
            # Python 3.7
            return typ.__origin__
        except AttributeError:
            return typ


def _get_object_class(theObj):
    return theObj.__class__.__qualname__


def _get_object_module(theObj):
    module = theObj.__class__.__module__
    if module == str.__class__.__module__:
        return None
    if module == "__main__":
        return getExecPath()
    return module


def _object_to_FQCN(theobj):
    """Gets module path of object"""
    module = theobj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return theobj.__class__.__qualname__
    return module + '.' + theobj.__class__.__qualname__
    #
    # return theobj.__module__ + '.' + theobj.__class__.__name__


def _find_class(moduleName, className):
    import importlib
    splitted_name = className.split(".")
    begin = True
    parentClass = None
    for name in splitted_name:
        if begin:
            parentClass = getattr(importlib.import_module(moduleName), name)
            begin = False
        else:
            parentClass = getattr(parentClass, name)
    return parentClass


def json_to_obj(json_dict):
    """Convenience class to create object from dictionary. Only works if CLASS_TAG is valid

    :param json_dict: dictionary loaded from a json file.
    :raise TypeError: if class can not be found
    :raise KeyError: if CLASS_TAG not present in dictionary
    """
    try:
        cls = _find_class(json_dict[MODULE_TAG], json_dict[CLASS_TAG])
        return json_to_obj_safe(json_dict, cls)
    except TypeError:
        raise
    except KeyError:
        raise KeyError("Object class not defined, can not create data. Use from_json_safe method to specify which class to load.")


def json_to_obj_safe(json_dict, cls):
    """Safe class to create object from dictionary.

    :param json_dict: dictionary loaded from a json file
    :param cls: class object to instantiate with dictionary
    """
    type_class = get_type_class(cls)

    if issubclass(type_class, list) or issubclass(type_class, tuple):
        list_type = cls.__args__[0]
        instance = list()
        for value in json_dict:
            instance.append(json_to_obj_safe(value, list_type))
        if issubclass(type_class, list):
            return instance
        else:
            return tuple(instance)
    elif issubclass(type_class, dict):
        instance = dict()

        for key, value in json_dict.items():
            key_type = cls.__args__[0]
            val_type = cls.__args__[1]

            # Check key
            if issubclass(key_type, str):
                key_instance = key
            else:
                key_instance = json_to_obj_safe(ast.literal_eval(key), key_type)

            # Check value
            try:
                isdict = issubclass(val_type, dict)
            except TypeError:
                isdict = False
            value_instance = value if isdict else json_to_obj_safe(value, val_type)

            # Updates
            instance.update({key_instance: value_instance})
        return instance
    elif issubclass(type_class, (float, int, complex, bool, str)):
        return json_dict
    elif issubclass(type_class, np.ndarray):
        return np.array(json_dict)
    else:  # Object must be instantiated
        theInstance = _instantiates_annotated_object(json_dict, cls)
        return theInstance


def _instantiates_annotated_object(_json_dict, _cls):
    # annotations: dict = _cls.__annotations__ if hasattr(_cls, '__annotations__') else None
    # Instantiate the object
    entranceParams = []
    try:
        params = inspect.signature(_cls).parameters
        for param in params.values():
            if param.default is param.empty:
                if param.annotation is not param.empty:
                    entranceParams.append(_instantiates_annotated_object({}, param.annotation))
                else:
                    entranceParams.append(None)
            else:
                entranceParams.append(param.default)
    except ValueError:
        pass

    try:
        instance = _cls(*entranceParams)
    except TypeError:
        printIfShown("Error while loading:\n{}".format(traceback.format_exc()), SHOW_ERROR)
        return None
    annotations = _get_annotations(_cls)

    # Set object attributes from data items

    # # Special care for option classes:
    # if issubclass(_cls, Option_class):
    #     # Take all attributes associated to options
    #     for additionalAttribute in Option_class.get_option_attributes():
    #         # If they have been saved (because options exist)
    #         if additionalAttribute in _json_dict:
    #             # Retrieve saved values
    #             value = _json_dict[additionalAttribute]
    #             field_type = annotations.get(additionalAttribute)
    #             # Set them
    #             rsetattr(instance, additionalAttribute, json_to_obj_safe(value, field_type))
    #     # Reinitialize instance. That will NOT reinitialize the options due to its "if not hasattr" in init:)
    #     instance.__init__(*entranceParams)

    # Special care for annotations

    if annotations:
        for name, value in _json_dict.items():
            if name not in EXCLUDED_TAGS:
                field_type = annotations.get(name)
                if _isclass(field_type) and isinstance(value, (dict, tuple, list, set, frozenset)):
                    try:
                        rsetattr(instance, name, json_to_obj_safe(value, field_type))
                    except AttributeError:
                        printIfShown("Failed to instantiate subfield **{}** with value **{}**".format(name, value, SHOW_WARNING))
                else:
                    rsetattr(instance, name, value)

    # Special care for Savable classes
    if isinstance(instance, SaveableObject):
        # Manage direct unknown attribute
        for additionalAttribute in instance.get_additional_attributes_to_save():
            try:
                theSubDict = _json_dict[additionalAttribute]

                if isinstance(theSubDict, (float, int, complex, bool, str)):  # Smallest quantity => use it has it is
                    rsetattr(instance, additionalAttribute, theSubDict)
                else:
                    try:
                        rsetattr(instance, additionalAttribute, json_to_obj(theSubDict))
                    except TypeError:
                        printIfShown("Failed to instantiate attribute **{}**".format(additionalAttribute, theSubDict), SHOW_WARNING)
            except KeyError:
                printIfShown("Attribute {} was not saved. Used default value.".format(additionalAttribute), SHOW_WARNING)
        # Manage list of unknown attributes
        for additionalAttribute in instance.get_additional_attributes_to_save_list():
            list_of_json = _json_dict[additionalAttribute]
            list_of_obj = list()
            for theSubDict in list_of_json:
                try:
                    list_of_obj.append(json_to_obj(theSubDict))
                except TypeError:
                    printIfShown("Failed to instantiate attribute **{}**".format(additionalAttribute, theSubDict), SHOW_WARNING)
            rsetattr(instance, additionalAttribute, list_of_obj)
    return instance


def _get_annotations(theObj):
    """Return annotated attributes (theObj being the type of the object)"""
    annotations: dict = theObj.__annotations__ if hasattr(theObj, '__annotations__') else dict()
    parents = inspect.getmro(theObj)[1:]  # Starting index 1 because 0 is self

    for parentClass in parents:
        if parentClass is not None and parentClass != object:
            theAnnotations = _get_annotations(parentClass)
            annotations.update(theAnnotations)
    return annotations


def obj_to_json(theObj):
    """Extract the json dictionary from the object. The data saved are automatically detected, using typehints.
     ex: x: int=5 will be saved, x=5 won't.
     Inheritance of annotation is managed by this function
     """

    def _to_json(recObj, is_first=False):
        """Use of is_first to save module class name"""
        if isinstance(recObj, list) or isinstance(recObj, tuple):
            theList = list()
            for elem in recObj:
                theList.append(_to_json(elem, is_first=is_first))
            return theList
        elif isinstance(recObj, dict):
            output_dict = dict()
            for key in list(recObj):
                output_dict[str(key)] = _to_json(recObj[key])
            return output_dict
        elif isinstance(recObj, int) or isinstance(recObj, float):
            return recObj
        elif isinstance(recObj, str):
            return recObj
        elif np.issubdtype(type(recObj), np.integer) or np.issubdtype(type(recObj), np.floating):
            return float(recObj)
        elif isinstance(recObj, bool):
            return str(recObj).lower()
        elif isinstance(recObj, np.ndarray):
            return _to_json(recObj.tolist())
        elif recObj is None:
            return "null"
        else:  # Item is a user-defined object
            output_dict = dict()
            if recObj is not None:
                if is_first:
                    output_dict[MODULE_TAG] = _get_object_module(recObj)
                    output_dict[CLASS_TAG] = _get_object_class(recObj)

                for attribute, is_first in _get_attributes_to_save(recObj):
                    try:
                        value = rgetattr(recObj, attribute)
                        output_dict[attribute] = _to_json(value, is_first=is_first)
                    except AttributeError:
                        pass
            return output_dict

    return _to_json(theObj, is_first=True)


def _get_attributes_to_save(theObj):
    """Return list (attribute, is_first)"""
    list_1 = list()
    list_2 = list()

    # List 1
    annotations = _get_annotations(type(theObj))
    if annotations is not None:
        list_1 = list(zip(annotations, [False]*len(annotations)))

    # List 2
    if isinstance(theObj, SaveableObject):
        additional_attributes = theObj.get_additional_attributes_to_save() + theObj.get_additional_attributes_to_save_list()
        list_2 = list(zip(additional_attributes, [True]*len(additional_attributes)))
    return list_1 + list_2


# def get_json_module_tree(theObj):
#     """Return dict containing {CLASS_TAG: "class_name", MODULE_TAG: "module_name", "attribute1":{"class_name": "module_name", ...}}"""
#     def _nested_module_tree(recObj):
#         theDict = dict()
#         theClass = _get_object_class(recObj)
#         theModule = _get_object_module(recObj)
#         if theModule is not None:
#             theDict[CLASS_TAG] = theClass
#             theDict[MODULE_TAG] = theModule
#         for attribute, _ in _get_attributes_to_save(recObj):
#             nestedObj = rgetattr(recObj, attribute)
#             nested_subtree = _nested_module_tree(nestedObj)
#             if nested_subtree:
#                 theDict[attribute] = nested_subtree
#         return theDict
#     return _nested_module_tree(theObj)


def get_json_module_tree_from_dict(jsonDict):
    """Return dict containing {CLASS_TAG: "class_name", MODULE_TAG: "module_name", "attribute1":{"class_name": "module_name", ...}}"""
    def _nested_module_tree(recDict):
        theDict = dict()
        theClass = recDict.get(CLASS_TAG)
        theModule = recDict.get(MODULE_TAG)
        if theModule is not None:
            theDict[CLASS_TAG] = theClass
            theDict[MODULE_TAG] = theModule
        for key in recDict:
            nestedObj = recDict[key]
            if isinstance(nestedObj, dict):
                nested_subtree = _nested_module_tree(nestedObj)
                if nested_subtree:
                    theDict[key] = nested_subtree
        return theDict
    return _nested_module_tree(jsonDict)


def remove_module_tree_from_string(theStr):
    """Used to compress string by removing __module__ and __class__ entries (used with get_json_module_tree_from_dict)"""
    str_copy = str(theStr)
    str_copy = re.sub("\s*\"__module__\" *: *(\"(.*?)\"(,|\s|)|\s*\{(.*?)\}(,|\s|))", '', str_copy)
    str_copy = re.sub("\s*\"__class__\" *: *(\"(.*?)\"(,|\s|)|\s*\{(.*?)\}(,|\s|))", '', str_copy)
    return str_copy


def apply_module_tree_to_dict(nestedTree, nestedObject, raiseError=False):
    """Restore __module__ and __class__ entries from nestedTree in nestedDict"""
    for key in nestedTree:
        if key == CLASS_TAG or key == MODULE_TAG:
            try:
                nestedObject[key] = nestedTree[key]
            except TypeError as e:
                printIfShown("(following error when loading, might come from numpy elementary types or 'null' values) {}".format(nestedTree, e), SHOW_WARNING)
                if raiseError:
                    raise
        else:
            if isinstance(nestedObject, dict):
                apply_module_tree_to_dict(nestedTree[key], nestedObject[key])


def encode_str_json(theStr):
    return theStr.__repr__().replace("\\", "\\\\")[1:-1]


def decode_str_json(theStr):
    return str(theStr.__repr__().replace("\\\\", "\\"))



from .onclickInterface import OnclickInterface


class Onclick_removeTrace(OnclickInterface):

    def __init__(self, theDataLink):
        """

        :param theDataLink: :class:`~optimeed.core.linkDataGraph.LinkDataGraph`
        """
        self.theDataLink = theDataLink

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, _):
        self.theDataLink.remove_collection(index_graph, index_trace)

    def get_name(self):
        return "Remove trace"



from .characterization import *
from .mathsToPhysics import *
from .objAndCons import *
from .optiAlgorithms import *
from .optiVariable import Real_OptimizationVariable, Binary_OptimizationVariable, Integer_OptimizationVariable
from .optimizer import run_optimization, OptimizerSettings
from .optiHistoric import OptiHistoric



from .onclickInterface import OnclickInterface
from optimeed.core import printIfShown, SHOW_WARNING
import traceback


class Onclick_animate(OnclickInterface):
    """On click: add or remove an element to animate"""

    def __init__(self, theLinkDataGraph, theAnimation):
        """

        :param theLinkDataGraph: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        :param theAnimation: :class:`~DataAnimationVisuals`
        """
        self.theLinkDataGraph = theLinkDataGraph
        self.theAnimation = theAnimation

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        try:
            trace_id = int(index_graph) * 100 + int(index_trace)
            theTrace = theGraphVisual.get_graph(index_graph).get_trace(index_trace)

            # index_in_data = theTrace.get_data().get_dataIndex_from_graphIndex(index_point)

            permutation = theTrace.theData.get_permutations()
            indicesInGraph = list(range(len(permutation)))  # Orders them accordingly to plot (data -> plot)
            all_listsOfDevices = self.theLinkDataGraph.get_clicked_items(index_graph, index_trace, indicesInGraph)

            # If trace wasn't present in any window ...
            if not self.theAnimation.contains_trace(trace_id):
                self.theAnimation.add_trace(trace_id, all_listsOfDevices, theTrace)
            for index_point in indices_points:
                index_in_graph = index_point
                self.theAnimation.add_elementToTrace(trace_id, index_in_graph)

            self.theAnimation.run()
        except KeyboardInterrupt:
            raise
        except Exception:
            printIfShown("Following error occurred in visualisation :" + traceback.format_exc(), SHOW_WARNING)

    def get_name(self):
        return "Animation"



import time
from multiprocessing import Pool, cpu_count

from optimeed.core.tools import printIfShown, SHOW_INFO, SHOW_WARNING, indentParagraph
from optimeed.optimize.optiVariable import Binary_OptimizationVariable, Integer_OptimizationVariable, Real_OptimizationVariable
from optimeed.core import Option_class, Option_int, Option_str, Option_dict
from .algorithmInterface import AlgorithmInterface
from .convergence import EvolutionaryConvergence
# Platypus imports
from .platypus import FixedLengthArray, Generator, Solution
from .platypus.algorithms import NSGAIII, SPEA2, SMPSO, NSGAII, OMOPSO
from .platypus.core import Problem, TerminationCondition, nondominated, unique
from .platypus.evaluator import Evaluator, run_job
from .platypus.operators import CompoundOperator, SBX, HUX, PM, BitFlip
from .platypus.types import Real, Integer, Binary


class MyProblem(Problem):
    """Automatically sets the optimization problem"""

    def __init__(self, theOptimizationVariables, nbr_objectives, nbr_constraints, evaluationFunction):
        super(MyProblem, self).__init__(len(theOptimizationVariables), nbr_objectives, nbr_constraints)

        # Convert types of optimization variables
        for i in range(len(self.types)):
            optimizationVariable = theOptimizationVariables[i]
            if type(optimizationVariable) is Real_OptimizationVariable:
                self.types[i] = Real(optimizationVariable.get_min_value(), optimizationVariable.get_max_value())
            elif type(optimizationVariable) is Integer_OptimizationVariable:
                self.types[i] = Integer(optimizationVariable.get_min_value(), optimizationVariable.get_max_value())
            elif type(optimizationVariable) is Binary_OptimizationVariable:
                self.types[i] = Binary(1)
            else:
                raise ValueError("Optimization variable not managed with this algorithm")

        self.evaluationFunction = evaluationFunction
        self.constraints[:] = "<=0"
        self.directions = FixedLengthArray(nbr_objectives, self.MINIMIZE)

    def evaluate(self, solution):
        x = solution.variables[:]
        returnedValues = self.evaluationFunction(x)

        for i in range(len(returnedValues["objectives"])):
            solution.objectives[i] = returnedValues["objectives"][i]
        for i in range(len(returnedValues["constraints"])):
            solution.constraints[i] = returnedValues["constraints"][i]

        solution.junk = returnedValues  # NEW VARIABLE ADDED


class MyGenerator(Generator):
    """Population generator to insert initial individual"""

    def __init__(self, initialVectorGuess):
        super(MyGenerator, self).__init__()
        self.initialVectorGuess = initialVectorGuess
        self.inserted_initialVector = False

    def generate(self, problem):
        solution = Solution(problem)
        if not self.inserted_initialVector:
            solution.variables = [x.encode(self.initialVectorGuess[i]) for i, x in enumerate(problem.types)]
            self.inserted_initialVector = True
        else:
            solution.variables = [x.rand() for x in problem.types]
        return solution


class MaxTimeTerminationCondition(TerminationCondition):
    def __init__(self, maxTime):
        super(MaxTimeTerminationCondition, self).__init__()
        self.maxTime = maxTime
        self.startingTime = None

    def initialize(self, algorithm):
        self.startingTime = time.time()

    def shouldTerminate(self, algorithm):
        return time.time() - self.startingTime > self.maxTime


class ConvergenceTerminationCondition(TerminationCondition):
    def __init__(self, minrelchange_percent=0.1, nb_generation=15):
        super(ConvergenceTerminationCondition, self).__init__()
        self.minrelchange = minrelchange_percent
        self.nb_generation = nb_generation

    def initialize(self, algorithm):
        pass

    def shouldTerminate(self, algorithm):
        convergence = algorithm.archive.conv
        if convergence.last_step() <= self.nb_generation:
            return False
        try:
            curr_hypervolume, _ = convergence.get_hypervolume(convergence.get_pareto_at_step(convergence.last_step()))
            last_hypervolume, _ = convergence.get_hypervolume(convergence.get_pareto_at_step(convergence.last_step() - self.nb_generation))
            rel_change = abs((curr_hypervolume - last_hypervolume)/curr_hypervolume * 100)  # percent
            printIfShown("Current hypervolume: {} Before hypervolume: {} Rel Change: {}".format(curr_hypervolume, last_hypervolume, rel_change), SHOW_INFO)
            if rel_change < self.minrelchange:
                printIfShown("terminating because converged !", SHOW_INFO)
            return rel_change < self.minrelchange
        except (IndexError, ZeroDivisionError):
            return False


class SeveralTerminationCondition(TerminationCondition):
    def __init__(self):
        super().__init__()
        self.listOfTerminationConditions = list()

    def initialize(self, algorithm):
        for terminationCondition in self.listOfTerminationConditions:
            terminationCondition.initialize(algorithm)

    def add(self, theTerminationCondition):
        if isinstance(theTerminationCondition, TerminationCondition):
            self.listOfTerminationConditions.append(theTerminationCondition)
        else:
            printIfShown("Invalid termination condition", SHOW_WARNING)

    def shouldTerminate(self, algorithm):
        return any([terminationCondition.shouldTerminate(algorithm) for terminationCondition in self.listOfTerminationConditions])


class MyMapEvaluator(Evaluator):
    def __init__(self, callback_on_evaluation):
        super().__init__()
        self.callback_on_evaluation = callback_on_evaluation

    def evaluate_all(self, jobs, **kwargs):
        outputs = [None] * len(jobs)
        for k, job in enumerate(jobs):
            outputs[k] = run_job(job)
            self.callback_on_evaluation(outputs[k].solution.junk)
            del outputs[k].solution.junk

        return outputs


class MyMultiprocessEvaluator(Evaluator):
    def __init__(self, callback_on_evaluation, numberOfCores):
        super().__init__()
        self.callback_on_evaluation = callback_on_evaluation
        self.pool = Pool(numberOfCores)

    def my_callback(self, output):
        self.callback_on_evaluation(output.solution.junk)
        del output.solution.junk

    def evaluate_all(self, jobs, **kwargs):
        outputs = [self.pool.apply_async(run_job, args=(job,), callback=self.my_callback) for job in jobs]
        return [output.get() for output in outputs]

    def close(self):
        printIfShown("Closing Pool", SHOW_INFO)
        self.pool.close()
        printIfShown("Waiting for all processes to complete", SHOW_INFO)
        self.pool.join()
        printIfShown("Pool closed", SHOW_INFO)


class MultiObjective_GA(AlgorithmInterface, Option_class):
    """Based on `Platypus Library <https://platypus.readthedocs.io/en/docs/index.html>`_.
    Workflow:
    Define what to optimize and which function to call with a :class:`Problem`
    Define the initial population with a :class:`Generator`
    Define the algorithm. As options, define how to evaluate the elements with a :class:`Evaluator`, i.e., for multiprocessing.
    Define what is the termination condition of the algorithm with :class:`TerminationCondition`. Here, termination condition is a maximum time.
    """
    DIVISION_OUTER = 0
    OPTI_ALGORITHM = 1
    NUMBER_OF_CORES = 2
    KWARGS_ALGO = 3

    def __init__(self):
        super().__init__()

        self.maxTime = None  # set by set_maxtime
        self.evaluationFunction = None  # set by set_max_objective
        self.callback_on_evaluation = None  # set by set_max_objective
        self.initialPopulation = None  # set by set_initial_population OR randomly generated (default)
        self.currentPopulation = None
        self.numberOfObjective = None
        self.numberOfConstraints = None
        self.theDimensionalConstraints = None

        self.terminationConditions = SeveralTerminationCondition()
        self.kwargs_opti_algorithm = dict()  # Additional kwargs opti_algorithm

        self.add_option(self.OPTI_ALGORITHM, Option_str("Optimization algorithm", 'NSGAII', choices=["NSGAII", "NSGAIII", "OMOPSO", "SPEA2", "SMPSO"]))
        self.add_option(self.NUMBER_OF_CORES, Option_int("Number of cores used in evaluation", 1))
        self.add_option(self.KWARGS_ALGO, Option_dict("Keywords arguments to send to the optimization algorithm", {}))

        self.algorithm = None

    def initialize(self, initialVectorGuess, listOfOptimizationVariables):
        """This function is called just before running optimization algorithm."""

        theProblem = MyProblem(listOfOptimizationVariables, self.numberOfObjective, self.numberOfConstraints, self.evaluationFunction)

        kwargs = {"generator": MyGenerator(initialVectorGuess), "convergence": EvolutionaryConvergence()}
        # Update evaluator for multiprocessing
        if self.get_option_value(self.NUMBER_OF_CORES) > 1:
            kwargs.update({"evaluator": MyMultiprocessEvaluator(callback_on_evaluation=self.callback_on_evaluation,
                                                                numberOfCores=min(cpu_count(), self.get_option_value(self.NUMBER_OF_CORES))
                                                                )})
        else:
            kwargs.update({"evaluator": MyMapEvaluator(callback_on_evaluation=self.callback_on_evaluation)})

        # Update variator if different types to optimize
        base_type = theProblem.types[0].__class__
        if not all([isinstance(t, base_type) for t in theProblem.types]):
            kwargs.update({"variator": CompoundOperator(SBX(), HUX(), PM(), BitFlip())})

        kwargs.update(self.get_option_value(self.KWARGS_ALGO))

        # Set optimization algorithm
        divisions_outer = int(30 * self.numberOfObjective)
        if self.get_option_value(self.OPTI_ALGORITHM) == 'SPEA2':
            algorithm = SPEA2(theProblem, **kwargs)
        elif self.get_option_value(self.OPTI_ALGORITHM) == 'SMPSO':
            algorithm = SMPSO(theProblem, **kwargs)
        elif self.get_option_value(self.OPTI_ALGORITHM) == 'OMOPSO':
            algorithm = OMOPSO(theProblem, epsilons=[0.05], **kwargs)
        elif self.get_option_value(self.OPTI_ALGORITHM) == 'NSGAII':
            algorithm = NSGAII(theProblem, **kwargs)  # self.get_option_value(self.DIVISION_OUTER))
        elif self.get_option_value(self.OPTI_ALGORITHM) == 'NSGAIII':
            algorithm = NSGAIII(theProblem, divisions_outer, **kwargs)  # self.get_option_value(self.DIVISION_OUTER))
        else:
            raise NotImplementedError("This algorithm has not been yet implemented {}".format(self.get_option_value(self.OPTI_ALGORITHM)))
        self.algorithm = algorithm

    def compute(self):
        self.terminationConditions.add(MaxTimeTerminationCondition(self.maxTime))
        self.algorithm.run(self.terminationConditions)

        self.algorithm.evaluator.close()

        def decode_solution_platypus(var_solutions, var_optimisations):
            my_solutions = [None] * len(var_solutions)
            for k, type_var in enumerate(var_optimisations):
                my_solutions[k] = type_var.decode(var_solutions[k])
            return my_solutions

        return [decode_solution_platypus(solution.variables, self.algorithm.problem.types) for solution in unique(nondominated(self.algorithm.result)) if solution.feasible]

    def set_evaluationFunction(self, evaluationFunction, callback_on_evaluation, numberOfObjectives, numberOfConstraints):
        self.evaluationFunction = evaluationFunction
        self.callback_on_evaluation = callback_on_evaluation
        self.numberOfObjective = numberOfObjectives
        self.numberOfConstraints = numberOfConstraints

    def set_maxtime(self, maxTime):
        self.maxTime = maxTime

    def __str__(self):
        theStr = ''
        theStr += "Platypus multiobjective library\n"
        theStr += indentParagraph(super().__str__(), indent_level=1)
        return theStr

    def get_convergence(self):
        """This function is called just before compute. Because the convergence is contained in opti algorithm, it must be created now."""
        return self.algorithm.convergence

    def add_terminationCondition(self, theTerminationCondition):
        self.terminationConditions.add(theTerminationCondition)

    def reset(self):
        terminationConditions = self.terminationConditions
        kwargs_opti_algorithm = self.kwargs_opti_algorithm
        super().reset()
        self.terminationConditions = terminationConditions
        self.kwargs_opti_algorithm = kwargs_opti_algorithm



from .onclickInterface import OnclickInterface
from .collectionExporterGUI import CollectionExporterGUI


class Onclick_exportCollection(OnclickInterface):
    """On click: export the selected points"""
    def __init__(self, theDataLink):
        """

        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        """
        self.theDataLink = theDataLink
        self.the_collection_exporter = CollectionExporterGUI()

        self.the_collection_exporter.signal_has_exported.connect(self.the_collection_exporter.reset)
        self.the_collection_exporter.signal_has_reset.connect(self.reset_graph)

        self.modifiedTraces = list()

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        theTrace = theGraphVisual.get_graph(index_graph).get_trace(index_trace)
        self.modifiedTraces.append(theTrace)
        # Add the point to the collection exporter
        for index_point in indices_points:
            theTrace.set_brush(index_point, (250, 250, 0))
            theData = self.theDataLink.get_clicked_item(index_graph, index_trace, index_point)
            self.the_collection_exporter.add_data_to_collection(theData)

    def reset_graph(self):
        for trace in self.modifiedTraces:
            trace.reset_all_brushes()
        self.modifiedTraces = list()

    def get_name(self):
        return "Select points to export collection"





import ast
import inspect
import operator
import os
import re
import sys

import numpy as np

from optimeed.core.commonImport import *

_workspace_path = os.path.abspath(os.getcwd() + '/Workspace')


class text_format:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    WHITE = '\033[30m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


def software_version():
    MAJOR = 1.0
    MINOR = 0.0
    return MAJOR * 10 + MINOR


# # Find an expression delimited by begin_char and end_char in theStr, and apply transformation replace_function
# ex: find_and_replace('{', '}', 2*{x}, lambda str_in: str_in + "*2") returns 2*x*2
def find_and_replace(begin_char, end_char, theStr, replace_function):
    regex = '\\' + begin_char + '([^}]+)' + '\\' + end_char
    split_str = re.split(regex, theStr)  # Replace chars are even
    new_str = ''
    for i in range(len(split_str)):
        if i % 2:
            new_str += str(replace_function(split_str[i]))
        else:
            new_str += str(split_str[i])
    return new_str


def create_unique_dirname(dirname):
    """Create dirname if it doesn't exists, otherwise append an integer to dirname and create it.

    :param dirname: name of the directory to create
    :return: name of the directory created"""
    base_dir = dirname
    index = 0
    while os.path.isdir(dirname):
        dirname = base_dir + "_" + str(index)
        index += 1
    os.makedirs(os.path.abspath(dirname))
    return dirname


def applyEquation(objectIn, s):
    """Apply literal expression based on an object

    :param objectIn: Object
    :param s: literal expression. Float variables taken from the object are written between {}, int between []. Example: s="{x}+{y}*2" if x and y are attributes of objectIn.
    :return: value (float)
    """
    str_eq = find_and_replace("{", "}", s, lambda attr_in: "({})".format(float(rgetattr(objectIn, attr_in))))
    str_eq = find_and_replace("[", "]", str_eq, lambda attr_in: "({})".format(int(rgetattr(objectIn, attr_in))))
    return eval(str_eq)


def arithmeticEval(s):
    node = ast.parse(s, mode='eval')
    binOps = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Mod: operator.mod,
        ast.USub: operator.neg
    }

    def _eval(in_node):
        if isinstance(in_node, ast.Expression):
            return _eval(in_node.body)
        elif isinstance(in_node, ast.Str):
            return in_node.s
        elif isinstance(in_node, ast.Num):
            return in_node.n
        elif isinstance(in_node, ast.BinOp):
            return binOps[type(in_node.op)](_eval(in_node.left), _eval(in_node.right))
        elif isinstance(in_node, ast.Name):
            return float('inf')
        elif isinstance(in_node, ast.UnaryOp):
            return binOps[type(in_node.op)](_eval(in_node.operand))
        else:
            raise TypeError('Unsupported type {}'.format(in_node))

    return _eval(node.body)


def isNonePrintMessage(theObject, theMessage, show_type=SHOW_INFO):
    if theObject is None:
        printIfShown(theMessage, show_type)
        return True
    return False


def getPath_workspace():
    """Get workspace path (i.e., location where optimeed files will be created). Create directory if doesn't exist."""
    if not os.path.exists(_workspace_path):
        os.makedirs(_workspace_path)
        printIfShown("Created workspace folder: {}".format(_workspace_path), SHOW_INFO)
    return _workspace_path


def setPath_workspace(thePath):
    """Set workspace path (i.e., location where optimeed files will be created)"""
    global _workspace_path
    _workspace_path = os.path.abspath(os.path.expandvars(os.path.expanduser(thePath)))
    printIfShown("Set workspace folder: {}".format(_workspace_path), SHOW_INFO)


def getLineInfo(lvl=1):
    filename = os.path.relpath(inspect.stack()[lvl][1], os.getcwd())
    line = inspect.stack()[lvl][2]
    method = inspect.stack()[lvl][3]
    return filename, method, line


def printIfShown(theStr, show_type=SHOW_DEBUG, isToPrint=True, appendTypeName=True, end='\n'):
    if isToPrint:
        try:
            sys.stdout.isatty()
            if show_type == SHOW_WARNING:
                colourBegin = text_format.YELLOW  # yellow
                if appendTypeName:
                    colourBegin += 'Warning: '
            elif show_type == SHOW_ERROR:
                colourBegin = text_format.RED
                if appendTypeName:
                    colourBegin += 'ERROR: '
            elif show_type == SHOW_INFO:
                colourBegin = text_format.BLUE
                if appendTypeName:
                    colourBegin += 'Info: '
            elif show_type == SHOW_DEBUG:
                colourBegin = text_format.WHITE
                if appendTypeName:
                    colourBegin += 'Debug: '
            else:
                colourBegin = ''
            colourEnd = '' if colourBegin == '' else '\033[0m'
        except AttributeError:  # Not the original stdout that has not implemented that function :) :) :)
            colourBegin = ''
            colourEnd = ''

        if show_type in SHOW_CURRENT:
            if SHOW_LOGS in SHOW_CURRENT:
                file, method, line = getLineInfo(2)
                logstr = "\t[File: {}, Line: {}, Method: {}]".format(file, line, method)
            else:
                logstr = ""
            print(colourBegin + str(theStr) + colourEnd + logstr, end=end)
    return


def universalPath(thePath):
    return os.path.abspath(thePath)


def add_suffix_to_path(thePath, suffix):
    head, tail = os.path.split(thePath)
    base_name, extension = os.path.splitext(tail)
    newPath = head + '/' + base_name + suffix + extension
    return newPath


def get_object_attrs(obj):
    try:
        return obj.__dict__
    except AttributeError:
        dico = {}
        try:
            for attr in obj.__slots__:
                dico[attr] = getattr(obj, attr)
        except AttributeError:
            pass
        # return {attr: getattr(obj, attr) for attr in obj.__slots__}
        return dico


def rsetattr(obj, attr, val):
    """setattr, but recursively. Works with list (i.e. theObj.myList[0].var_x)"""
    pre, _, post = attr.rpartition('.')  # ex: "a.b.c" -> pre = a.b, ; post = c

    if pre:
        obj_returned = rgetattr(obj, pre)  # , checkIndices=False)
    else:
        obj_returned = obj

    splitted = post.split('[')
    index_end_list = []
    char_begin = attr
    if len(splitted) > 1:
        char_begin = splitted[0]
        for i in range(1, len(splitted)):
            index_end_list.append(splitted[i].replace(']', '').replace("'", ''))

    if len(index_end_list):
        curr_nested_obj = obj_returned
        for index in range(len(index_end_list) - 1):
            try:
                curr_nested_obj = curr_nested_obj[index_end_list[index]]
            except (KeyError, TypeError, IndexError):
                curr_nested_obj = curr_nested_obj[int(index_end_list[index])]

        try:
            getattr(curr_nested_obj, char_begin)[index_end_list[-1]] = val
        except (KeyError, TypeError, IndexError):
            getattr(curr_nested_obj, char_begin)[int(index_end_list[-1])] = val
    else:
        setattr(obj_returned, post, val)


def rgetattr(obj, attr):
    """getattr, but recursively. Works with list."""
    attributes = attr.split('.')
    curr_object = obj
    for attribute in attributes:
        # Get end parentheses
        list_indices = []
        splitted = attribute.split('[')
        if not splitted:
            splitted = [attribute]

        curr_object = getattr(curr_object, splitted[0])
        if len(splitted) > 1:
            for i in range(1, len(splitted)):
                list_indices.append(splitted[i].replace(']', '').replace("'", ''))

        for item in list_indices:
            try:
                curr_object = curr_object[item]
            except (KeyError, TypeError, IndexError):
                curr_object = curr_object[int(item)]
    return curr_object


def indentParagraph(text_in, indent_level=1):
    """Add '\t' at beginning of strings and after each '\n'."""
    rowsP = text_in.split("\n")
    if rowsP[-1] == '':
        rowsP = rowsP[0:-1]
    Pout = ''
    space = ''
    for i in range(indent_level):
        space += '\t'
    for element in rowsP:
        Pout += space + element + '\n'
    return Pout


# def from_scattered_to_surf(dataX, dataY, dataZ, nPoints):
#     X = np.arange(min(dataX), max(dataX), (max(dataX)-min(dataX))/nPoints)
#     Y = np.arange(min(dataY), max(dataY), (max(dataY)-min(dataY))/nPoints)
#     X, Y = np.meshgrid(X, Y)
#     Z = griddata((dataX, dataY), dataZ, (X, Y), method='linear')
#     return X, Y, Z


def truncate(theStr, truncsize):
    return theStr[:truncsize] + (theStr[truncsize:] and '..')


def get_recursive_attrs(theObject, max_recursion_level=2):
    def _all_attr(_theObject, _max_recursion_level, curr_level, prefix=""):
        temp = vars(_theObject)
        list_attributes = []
        object_keys = list()
        variable_keys = list()
        list_keys = list()

        for _attribute in temp:
            _theObject = temp[_attribute]
            if getattr(_theObject, '__dict__', None) is not None:
                object_keys.append(_attribute)
            else:
                if isinstance(_theObject, list):
                    list_keys.append(_attribute)
                else:
                    variable_keys.append(_attribute)
        for _attribute in variable_keys:
            list_attributes.append("{}{}".format(prefix, _attribute))

        for _attribute in list_keys:
            theList = temp[_attribute]

            for i, item in enumerate(theList):
                prefix_list = "{}{}[{}]".format(prefix, _attribute, i)
                list_attributes.append(prefix_list)

                if getattr(item, '__dict__', None) is not None:  # Object
                    list_attributes += _all_attr(item, _max_recursion_level, curr_level + 1, prefix="{}.".format(prefix_list))

        if curr_level >= _max_recursion_level:
            return list_attributes
        else:
            for key in object_keys:
                _theObject = temp[key]
                list_attributes += _all_attr(_theObject, _max_recursion_level, curr_level + 1, prefix="{}.".format(key))
        return list_attributes
    return _all_attr(theObject, max_recursion_level, 0)


def str_all_attr(theObject, max_recursion_level):
    def set_attribute_color(strAttribute, level):
        if level == 0:
            color = text_format.BLUE + text_format.BOLD
        elif level == 1:
            color = text_format.GREEN
        else:
            color = text_format.DARKCYAN
        return color + strAttribute + text_format.END

    def _str_all_attr(_theObject, _max_recursion_level, curr_level):
        temp = vars(_theObject)
        theStr = ''
        object_keys = list()
        variable_keys = list()
        list_keys = list()
        truncsize = 99999999999999999999
        for key in temp:
            _theObject = temp[key]
            if getattr(_theObject, '__dict__', None) is not None:
                object_keys.append(key)
            else:
                if isinstance(_theObject, list):
                    list_keys.append(key)
                else:
                    variable_keys.append(key)
        for key in variable_keys:
            theVariable = temp[key]
            theStr += set_attribute_color(key, curr_level) + ' : ' + set_attribute_color(truncate(str(theVariable), truncsize), curr_level) + '\n'

        for key in list_keys:
            theList = temp[key]
            theStr += set_attribute_color(key, curr_level) + '\t:\t'
            for index, item in enumerate(theList):
                if getattr(item, '__dict__', None) is not None:
                    if index == 0:
                        theStr += '\n'
                    theStr += indentParagraph("Item " + str(index) + " -", 1)
                    theStr += indentParagraph(_str_all_attr(item, _max_recursion_level, curr_level + 1), 2)
                else:
                    theStr += truncate(str(item), truncsize) + ', '
                    if theStr.endswith('.., '):
                        break

            theStr += '\n'

        if curr_level >= _max_recursion_level:
            return theStr
        else:
            for key in object_keys:
                _theObject = temp[key]
                theStr += set_attribute_color(key, curr_level) + ":\n"
                theStr += indentParagraph(_str_all_attr(_theObject, _max_recursion_level, curr_level + 1), 1)
        return theStr

    return _str_all_attr(theObject, max_recursion_level, 0)


def get_2D_pareto(xList, yList, max_X=True, max_Y=True):
    """Get 2D pareto front

    :param xList: list of x coordinates
    :param yList: list of y coordinates
    :param max_X: True if x is to maximize
    :param max_Y: true if y is to maximize
    :return: x pareto-optimal coordinates, y pareto-optimal coordinates, indices of these points in input parameters
    """
    coef_x = 1
    coef_y = 1
    if max_X:
        coef_x = -1
    if max_Y:
        coef_y = -1
    # assign all in new array and ignore values
    mapping = []
    coords = []
    for i in range(len(xList)):
        if not np.isnan(xList[i]) and not np.isnan(yList[i]):
            coords.append([coef_x * xList[i], coef_y * yList[i]])
            mapping.append(i)

    costs = np.array(coords)
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(costs[is_efficient] < c, axis=1)  # Keep any point with a lower cost
            is_efficient[i] = True  # And keep self
    # Create new array
    newx = []
    newy = []
    indices = []
    for index, isPareto in enumerate(is_efficient):
        if isPareto:
            goodIndex = mapping[index]
            newx.append(xList[goodIndex])
            newy.append(yList[goodIndex])
            indices.append(goodIndex)
    return newx, newy, indices


def get_ND_pareto(objectives_list, are_maxobjectives_list=None):
    """
    Return the N-D pareto front

    :param objectives_list: list of list of objectives: example [[0,1], [1,1], [2,2]]
    :param are_maxobjectives_list: for each objective, tells if they are to be maximized or not: example [True, False]. Default: False
    :return: extracted_pareto, indices: list of [x, y, ...] points forming the pareto front, and list of the indices of these points from the base list.
    """
    numberOfObjectives = len(objectives_list[0])
    if are_maxobjectives_list is None:
        are_maxobjectives_list = [False] * numberOfObjectives

    coef_list = [-1 if isMaxObjective else 1 for isMaxObjective in are_maxobjectives_list]

    # assign all in new array and ignore values
    mapping = []
    coords = []

    for index_objective, objectives in enumerate(objectives_list):
        mustAdd = True
        for objective in objectives:
            if np.isnan(objective):
                mustAdd = False
        if mustAdd:
            coords.append([coef_list[i] * objectives[i] for i in range(numberOfObjectives)])
            mapping.append(index_objective)

    costs = np.array(coords)
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(costs[is_efficient] < c, axis=1)  # Keep any point with a lower cost
            is_efficient[i] = True  # And keep self
    # Create new array
    extracted_pareto = list()
    indices = []
    for index, isPareto in enumerate(is_efficient):
        if isPareto:
            goodIndex = mapping[index]
            extracted_pareto.append(objectives_list[goodIndex])
            indices.append(goodIndex)
    return extracted_pareto, indices


def delete_indices_from_list(indices, theList):
    """Delete elements from list at indices

    :param indices: list
    :param theList: list
    """
    sorted_indices = list(dict.fromkeys(list(np.sort(indices))))
    sorted_indices.reverse()
    for index in sorted_indices:
        del theList[index]


def merge_two_dicts(dict1, dict2):
    """Merge two dicts without affecting them

    :return: new dictionary"""
    z = dict1.copy()
    z.update(dict2)
    return z


def deep_sizeof(obj):
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        return size + sum(map(deep_sizeof, obj.keys())) + sum(map(deep_sizeof, obj.values()))
    if isinstance(obj, (list, tuple, set, frozenset)):
        return size + sum(map(deep_sizeof, obj))
    return size


def order_lists(ref_list, linked_list):
    return map(list, zip(*sorted(zip(ref_list, linked_list), key=lambda x: x[0])))




from .onclickInterface import OnclickInterface
from PyQt5 import QtWidgets
import os
from optimeed.core import rgetattr


class Onclick_exportToTxt(OnclickInterface):
    """On click: export the data of the whole the trace selected"""
    def __init__(self, theDataLink, attributes_shadow=None):
        """

        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        :param attributes_shadow: list of attributes (as string) to recover from the shadow (=data points/devices)
        """
        self.theDataLink = theDataLink
        self.attributes_shadow = attributes_shadow if attributes_shadow is not None else list()

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        howToPlot = self.theDataLink.get_howToPlotGraph(index_graph)
        attributes = [howToPlot.attribute_x, howToPlot.attribute_y]

        # Write header
        theStr = ""
        for attribute in attributes:
            theStr += "{}\t".format(attribute)
        for attribute in self.attributes_shadow:
            theStr += "{}\t".format(attribute)
        theStr += "\n"

        # Recover data to export
        collection_displayed = self.theDataLink.get_collection_from_graph(index_graph, index_trace, getShadow=False)
        collection_shadowed = self.theDataLink.get_collection_from_graph(index_graph, index_trace, getShadow=True)

        for index_point, _ in enumerate(collection_displayed.get_data_generator()):
            dataObj = collection_displayed.get_data_at_index(index_point)
            for attribute in attributes:
                theStr += "{}\t".format(rgetattr(dataObj, attribute))

            if len(self.attributes_shadow):
                dataObj = collection_shadowed.get_data_at_index(index_point)
                for attribute in self.attributes_shadow:
                    theStr += "{}\t".format(rgetattr(dataObj, attribute))
            theStr += "\n"

        # Export
        dlg = QtWidgets.QFileDialog.getSaveFileName()[0]
        if dlg:
            root, ext = os.path.splitext(dlg)
            filename = root + ".txt"
            with open(filename, "w") as f:
                f.write(theStr)

    def get_name(self):
        return "Export trace to text file"



from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSlot


class Widget_listWithSearch(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout_widget = QtWidgets.QVBoxLayout(self)
        layout_widget.setContentsMargins(0, 0, 0, 0)

        self.myListWidget = QtWidgets.QListWidget()

        self.search_plot_data = QtWidgets.QLineEdit()
        # self.search_plot_data.setTabChangesFocus(True)

        self.layout().addWidget(self.search_plot_data)
        self.layout().addWidget(self.myListWidget)

        self.search_plot_data.textChanged.connect(self._filter_list)

    @pyqtSlot()
    def get_index_selected(self):
        return self.myListWidget.currentRow()

    @pyqtSlot()
    def get_name_selected(self):
        return self.myListWidget.currentItem().text()

    def set_list(self, names):
        for name in names:
            self.myListWidget.addItem(name)

    def _filter_list(self):
        text = self.search_plot_data.text()
        for item in self._iter_items():
            if text in item.text():
                item.setHidden(False)
            else:
                item.setHidden(True)
        # print(text)

    def _iter_items(self):
        for i in range(self.myListWidget.count()):
            yield self.myListWidget.item(i)



import shutil

from SALib.analyze import sobol
from optimeed.core import SaveableObject, AutosaveStruct, create_unique_dirname, SHOW_INFO, order_lists, printIfShown, Performance_ListDataStruct
from typing import List
from .sensitivity_analysis_evaluation import evaluate
from optimeed.core import SingleObjectSaveLoad
import numpy as np
from optimeed.core import getPath_workspace, SHOW_ERROR
from optimeed.optimize import Real_OptimizationVariable
from multiprocessing import Pool

import math
import os


_filename_sensitivityparams = "sensitivity_params.json"
_foldername_embarrassingly_parallel_results = "_jobs_results"
_filename_sensitivityresults = "sensitivity.json"


class SensitivityResults(SaveableObject):
    paramsToEvaluate: List[float]
    success: bool
    index: int

    def __init__(self):
        self.paramsToEvaluate = [0.0]
        self.device = None
        self.success = False
        self.index = 0

    def add_data(self, params, device, success, index):
        self.device = device
        self.success = success
        self.paramsToEvaluate = params
        self.index = index

    def get_additional_attributes_to_save(self):
        return ["device"]


class SensitivityParameters(SaveableObject):
    list_of_optimization_variables: List[Real_OptimizationVariable]
    param_values: List[List[float]]

    def __init__(self, param_values, list_of_optimization_variables, theDevice, theMathsToPhys, theCharacterization):
        """
        These are the parameters t
        :param list_of_optimization_variables: list of OptiVariables that are analyzed
        :param theDevice: /
        :param theMathsToPhys: /
        :param theCharacterization: /"""
        self.theDevice = theDevice
        self.theMathsToPhys = theMathsToPhys
        self.theCharacterization = theCharacterization
        self.list_of_optimization_variables = list_of_optimization_variables
        self.param_values = param_values

    def get_device(self):
        return self.theDevice

    def get_M2P(self):
        return self.theMathsToPhys

    def get_charac(self):
        return self.theCharacterization

    def get_optivariables(self):
        return self.list_of_optimization_variables

    def get_paramvalues(self):
        return self.param_values

    def get_additional_attributes_to_save(self):
        return ["theDevice", "theMathsToPhys", "theCharacterization"]


def get_sensitivity_problem(list_of_optimization_variables):
    """
    This is the first method to use. Convert a list of optimization varisbles to a SALib problem

    :param list_of_optimization_variables: List of optimization variables
    :return: SALib problem
    """
    num_vars = len(list_of_optimization_variables)
    names = list()
    bounds = list()

    for variable in list_of_optimization_variables:
        if isinstance(variable, Real_OptimizationVariable):
            names.append(variable.get_attribute_name())
            bounds.append([variable.get_min_value(), variable.get_max_value()])
        else:
            raise TypeError("Optimization variable must be of real type to perform this analysis")
    problem = {'num_vars': num_vars, 'names': names, 'bounds': bounds}
    return problem


def _get_sensitivity_result(output):
    """Convert output of "evaluate" function to SensitivityResult"""
    result = SensitivityResults()
    result.add_data(output["x"], output["device"], output["success"], output["index"])
    return result


def _get_job_args(theSensitivityParameters, index):
    """Convert sensitivityparameters at index to args used in "evaluate" function"""
    return [theSensitivityParameters.get_paramvalues()[index], theSensitivityParameters.get_device(),
            theSensitivityParameters.get_M2P(), theSensitivityParameters.get_charac(), theSensitivityParameters.get_optivariables(), index]


def _find_missings(theSensitivityParameters, studyname):
    missings = list()
    for index, _ in enumerate(theSensitivityParameters.get_paramvalues()):
        saved_filename = os.path.join(getPath_workspace(), studyname, _foldername_embarrassingly_parallel_results, "{}.json".format(index))
        if not os.path.exists(saved_filename):
            missings.append(index)
    return missings


def prepare_embarrassingly_parallel_sensitivity(theSensitivityParameters, studyname):
    project_foldername = os.path.join(getPath_workspace(), studyname)
    foldername_tempfiles = os.path.join(project_foldername, _foldername_embarrassingly_parallel_results)
    shutil.rmtree(foldername_tempfiles, ignore_errors=True)
    os.makedirs(foldername_tempfiles)  # Also create project_foldername dir
    SingleObjectSaveLoad.save(theSensitivityParameters, os.path.join(project_foldername, _filename_sensitivityparams))
    printIfShown("Files created. There will be {} indices to evaluate".format(len(theSensitivityParameters.get_paramvalues())), SHOW_INFO)


def launch_embarrassingly_parallel_sensitivity(theSensitivityParameters, studyname, index):
    saved_filename = os.path.join(getPath_workspace(), studyname, _foldername_embarrassingly_parallel_results, "{}.json".format(index))
    if not os.path.exists(saved_filename):
        output = evaluate(_get_job_args(theSensitivityParameters, index))
        result = _get_sensitivity_result(output)
        SingleObjectSaveLoad.save(result, saved_filename)


def gather_embarrassingly_parallel_sensitivity(theSensitivityParameters, studyname):
    missings = _find_missings(theSensitivityParameters, studyname)
    if len(missings):
        printIfShown("Could not gather results yet, several parameters remain unevaluated:", SHOW_ERROR)
        printIfShown("{}".format(missings), SHOW_ERROR)
        exit(-1)

    results = Performance_ListDataStruct()

    for index, _ in enumerate(theSensitivityParameters.get_paramvalues()):
        saved_filename = os.path.join(getPath_workspace(), studyname, _foldername_embarrassingly_parallel_results, "{}.json".format(index))
        with open(saved_filename, 'r') as f:
            theStr = f.read()
        results.add_json_data(theStr)
    results.save(os.path.join(getPath_workspace(), studyname, _filename_sensitivityresults))


def evaluate_sensitivities(theSensitivityParameters: SensitivityParameters,
                           numberOfCores=2, studyname="sensitivity", indices_to_evaluate=None):
    """
    Evaluate the sensitivities

    :param theSensitivityParameters: class`~SensitivityParameters`
    :param numberOfCores: number of core for multicore evaluation
    :param studyname: Name of the study, that will be the subfolder name in workspace
    :param indices_to_evaluate: if None, evaluate all param_values, otherwise if list: evaluate subset of param_values defined by indices_to_evaluate
    :return: collection of class`~SensitivityResults`
    """
    myDataStruct = Performance_ListDataStruct()
    foldername = create_unique_dirname(os.path.join(getPath_workspace(), studyname))

    SingleObjectSaveLoad.save(theSensitivityParameters, os.path.join(foldername, _filename_sensitivityparams))
    # Start saving
    autosaveStruct = AutosaveStruct(myDataStruct, filename=os.path.join(foldername, _filename_sensitivityresults))
    autosaveStruct.start_autosave(timer_autosave=60*5)

    param_values = theSensitivityParameters.get_paramvalues()
    try:
        param_values = param_values.tolist()
    except AttributeError:
        pass

    if indices_to_evaluate is None:
        indices = list(range(len(param_values)))
    else:
        indices = indices_to_evaluate
        param_values = [param_values[index] for index in indices_to_evaluate]

    # create jobs
    jobs = [_get_job_args(theSensitivityParameters, index) for index in indices]

    pool = Pool(numberOfCores)
    nb_to_do = len(param_values)
    nb_done = 0
    permutations= list()
    for output in pool.imap_unordered(evaluate, jobs):
        result = _get_sensitivity_result(output)
        myDataStruct.add_data(result)
        permutations.append(output["index"])
        nb_done += 1
        printIfShown("did {} over {}".format(nb_done, nb_to_do), SHOW_INFO)

    # save results
    autosaveStruct.stop_autosave()

    myDataStruct.reorder(permutations=permutations)
    autosaveStruct.save()

    pool.close()
    pool.join()
    return myDataStruct


def analyse_sobol_create_array(theSensitivityParameters: SensitivityParameters, objectives):
    """
    Create readible result array, ordered by decreasing sobol indices.

    :param theSensitivityParameters: class:`SensitivityParameters`
    :param objectives: array-like of objective
    :return: tuples of STR, for S1 and ST
    """
    problem_SALib = get_sensitivity_problem(theSensitivityParameters.get_optivariables())
    Si = sobol.analyze(problem_SALib, np.array(objectives))

    nb_params = len(theSensitivityParameters.get_optivariables())
    _, ordered_S1 = order_lists(Si['S1'], list(range(nb_params)))
    _, ordered_ST = order_lists(Si['ST'], list(range(nb_params)))
    ordered_S1.reverse()
    ordered_ST.reverse()

    def format_array(indices, prefix):
        theStr = ''
        theStr += '─' * 120 + '\n'
        theStr += "{:^12}{:^14}{:^25}{:<}".format(*["Rank (" + prefix + ")", "Sobol value", "+- 95% conf", "Param name"]) + '\n'
        theStr += '─' * 120 + '\n'
        for i, map_index in enumerate(indices):
            row = [i + 1, Si[prefix][map_index], Si[prefix + '_conf'][map_index], theSensitivityParameters.get_optivariables()[map_index].get_attribute_name()]
            theStr += "{:^12}{:^14.3f}{:^25.3f}{}".format(*row) + '\n'
        theStr += '─' * 50 + '\n'
        theStr += "{:^12}{:^14.3f}{:^25}{:<}".format("SUM", sum(Si[prefix]), "", "") + '\n'
        return theStr

    S1_array = format_array(ordered_S1, "S1")
    ST_array = format_array(ordered_ST, "ST")
    print(S1_array)
    print(ST_array)
    return S1_array, ST_array


def analyse_sobol_convergence(theSensitivityParameters: SensitivityParameters, objectives, stepsize=1):
    """
    Create dictionary for convergence plot

    :param theSensitivityParameters: class:`SensitivityParameters`
    :param objectives: array-like of objective
    :return: Dictionary
    """
    problem_SALib = get_sensitivity_problem(theSensitivityParameters.get_optivariables())
    opti_variables = theSensitivityParameters.get_optivariables()
    nb_params = len(opti_variables)

    max_nb_step = math.floor(len(objectives) / (2*nb_params + 2))

    outputs = list()
    steps = list(range(1, max_nb_step, stepsize))
    for sample_size in steps:
        printIfShown("Doing {} over {}".format(sample_size, max_nb_step))
        outputs.append(sobol.analyze(problem_SALib, np.array(objectives[0:sample_size*(2*nb_params+2)])))

    outputs_dict = dict()
    for i in range(nb_params):
        outputs_dict[i] = {'S1': [output['S1'][i] for output in outputs],
                           'ST': [output['ST'][i] for output in outputs],
                           'step': steps,
                           'name': opti_variables[i].get_attribute_name()}
    return outputs_dict





# Copyright 2015-2018 David Hadka
#
# This file is part of Platypus, a Python module for designing and using
# evolutionary algorithms (EAs) and multiobjective evolutionary algorithms
# (MOEAs).
#
# Platypus is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Platypus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Platypus.  If not, see <http://www.gnu.org/licenses/>.
import unittest
import functools
from ..tools import *
from .test_core import createSolution


class TestBinary(unittest.TestCase):
    
    EXPECTED = {
        0 : { "binary" : (0, 0, 0, 0), "gray" : (0, 0, 0, 0)},
        1 : { "binary" : (0, 0, 0, 1), "gray" : (0, 0, 0, 1)},
        2 : { "binary" : (0, 0, 1, 0), "gray" : (0, 0, 1, 1)},
        3 : { "binary" : (0, 0, 1, 1), "gray" : (0, 0, 1, 0)},
        4 : { "binary" : (0, 1, 0, 0), "gray" : (0, 1, 1, 0)},
        5 : { "binary" : (0, 1, 0, 1), "gray" : (0, 1, 1, 1)},
        6 : { "binary" : (0, 1, 1, 0), "gray" : (0, 1, 0, 1)},
        7 : { "binary" : (0, 1, 1, 1), "gray" : (0, 1, 0, 0)},
        8 : { "binary" : (1, 0, 0, 0), "gray" : (1, 1, 0, 0)},
        9 : { "binary" : (1, 0, 0, 1), "gray" : (1, 1, 0, 1)},
        10 : { "binary" : (1, 0, 1, 0), "gray" : (1, 1, 1, 1)},
        11 : { "binary" : (1, 0, 1, 1), "gray" : (1, 1, 1, 0)},
        12 : { "binary" : (1, 1, 0, 0), "gray" : (1, 0, 1, 0)},
        13 : { "binary" : (1, 1, 0, 1), "gray" : (1, 0, 1, 1)},
        14 : { "binary" : (1, 1, 1, 0), "gray" : (1, 0, 0, 1)},
        15 : { "binary" : (1, 1, 1, 1), "gray" : (1, 0, 0, 0)},
    }
    
    def assertBinEqual(self, b1, b2):
        self.assertEqual(len(b1), len(b2))
        
        for i in range(len(b1)):
            self.assertEqual(bool(b1[i]), bool(b2[i]))
    
    def test_int2bin(self):
        self.assertBinEqual([], int2bin(0, 0))
        self.assertBinEqual([0], int2bin(0, 1))
        
        for i in range(16):
            self.assertBinEqual(self.EXPECTED[i]["binary"], int2bin(i, 4))
        
    def test_bin2int(self):
        self.assertEqual(0, bin2int([]))
        self.assertEqual(0, bin2int([0]))

        for i in range(16):
            self.assertEqual(i, bin2int(self.EXPECTED[i]["binary"]))
        
    def test_bin2gray(self):
        for i in range(16):
            self.assertBinEqual(self.EXPECTED[i]["gray"], bin2gray(int2bin(i, 4)))
        
    def test_gray2bin(self):
        for i in range(16):
            self.assertBinEqual(self.EXPECTED[i]["binary"], gray2bin(self.EXPECTED[i]["gray"]))
            

class TestVectorAlgebra(unittest.TestCase):
    
    def test_dot(self):
        self.assertEqual(0, dot([0, 0, 0], [1, 1, 1]))
        self.assertEqual(1, dot([1, 0, 1], [0, 1, 1]))
        self.assertEqual(2, dot([1, 0, 1], [1, 1, 1]))
        
    def test_subtract(self):
        self.assertEquals([1, 0, -1], subtract([1, 1, 1], [0, 1, 2]))
        
    def test_multiply(self):
        self.assertEquals([0, 0, 0], multiply(0, [1, 1, 1]))
        self.assertEquals([.5, .5, .5], multiply(.5, [1, 1, 1]))
        
    def test_magnitude(self):
        self.assertEqual(0, magnitude([0, 0, 0]))
        self.assertAlmostEqual(1, magnitude([0.577, 0.577, 0.577]), delta=0.001)
    
    def test_distance1(self):
        line = [1, 1, 1]
        point = [2, 2, 2]
        self.assertEqual(0.0, point_line_dist(point, line))
        
    def test_distance2(self):
        line = [1, 1, 1]
        point = [0, 0, 1]
        self.assertAlmostEqual(0.816, point_line_dist(point, line), places=3)
        
    def test_distance3(self):
        line = [1, 0, 0]
        point = [0, 0, 1]
        self.assertEqual(1.0, point_line_dist(point, line))
        
    def test_lsolve1(self):
        A = [[1, 0], [0, 1]]
        b = [1, 1]
        self.assertEquals([1, 1], lsolve(A, b))
        
    def test_lsolve2(self):
        A = [[0.7, 0.3, 0.0], [0.1, 0.1, 0.7], [0.2, 0.1, 0.9]]
        b = [0.2, 0.3, 0.5]
        x = lsolve(A, b)
        self.assertAlmostEqual(0.965, x[0], delta=0.001)
        self.assertAlmostEqual(-1.586, x[1], delta=0.001)
        self.assertAlmostEqual(0.517, x[2], delta=0.001)
        
    def test_lsolve_singular(self):
        A = [[0.5, 0.5], [0.5, 0.5]]
        b = [1, 1]
        self.assertRaises(SingularError, functools.partial(lsolve, A, b))

class TestCombinatorics(unittest.TestCase):
    
    def test_choose(self):
        self.assertEqual(1, choose(0, 0))
        self.assertEqual(1, choose(1, 0))
        self.assertEqual(1, choose(5, 0))
        self.assertEqual(1, choose(1, 1))
        self.assertEqual(5, choose(5, 1))
        self.assertEqual(5, choose(5, 4))
        self.assertEqual(1, choose(5, 5))
        
class TestDistances(unittest.TestCase):
    
    def test_euclidean(self):
        self.assertEqual(0.0, euclidean_dist([1, 1], [1, 1]))
        self.assertAlmostEqual(1.414, euclidean_dist([0, 0], [1, 1]), delta=0.001)
        self.assertAlmostEqual(1.414, euclidean_dist([1, 1], [0, 0]), delta=0.001)
        
class TestDistanceMatrix(unittest.TestCase):
    
    def test(self):
        solutions = [createSolution(0, 1), createSolution(0.5, 0.5), createSolution(0.75, 0.25), createSolution(1, 0)]
        matrix = DistanceMatrix(solutions)
        
        self.assertAlmostEqual(0.353, matrix[1, 2], delta=0.001)
        self.assertAlmostEqual(0.353, matrix[2, 1], delta=0.001)
        self.assertAlmostEqual(0.353, matrix.kth_distance(2, 0), delta=0.001)
        self.assertAlmostEqual(0.353, matrix.kth_distance(1, 0), delta=0.001)
        self.assertEqual(2, matrix.find_most_crowded())
        matrix.remove_point(2)
        self.assertEqual(1, matrix.find_most_crowded())

class TestDictMethods(unittest.TestCase):

    def test_remove_keys(self):
        self.assertEqual({}, remove_keys({}))
        self.assertEqual({}, remove_keys({}, "a", "b"))
        self.assertEqual({}, remove_keys({"a" : "remove"}, "a", "b"))
        self.assertEqual({"c" : "keep"}, remove_keys({"a" : "remove", "c" : "keep"}, "a", "b"))
        self.assertEqual({"a" : "keep"}, remove_keys({"a" : "keep"}))

    def test_only_keys(self):
        self.assertEqual({}, only_keys({}))
        self.assertEqual({}, only_keys({}, "a", "b"))
        self.assertEqual({"a" : "keep"}, only_keys({"a" : "keep", "b" : "remove"}, "a"))

    def _test_func_pos(self, a):
        pass

    def _test_func_def(self, a=5):
        pass

    def test_keys_for(self):
        self.assertEqual({}, only_keys_for({}, self._test_func_pos))
        self.assertEqual({"a" : "keep"}, only_keys_for({"a" : "keep", "b" : "remove"}, self._test_func_pos))
        self.assertEqual({}, only_keys_for({}, self._test_func_def))
        self.assertEqual({"a" : "keep"}, only_keys_for({"a" : "keep", "b" : "remove"}, self._test_func_def))


from .widget_listWithSearch import Widget_listWithSearch
from .widget_image import Widget_image
from .widget_lineDrawer import Widget_lineDrawer
from .widget_listWithSearch import Widget_listWithSearch
from .widget_menuButton import Widget_menuButton
from .widget_openGL import Widget_openGL
from .widget_tableWithSearch import Widget_tableWithSearch
from .widget_text import Widget_text, Widget_text_scrollable
from .openGL import *


import matplotlib.pyplot as plt
from PyQt5 import QtWidgets, uic
from optimeed.core.tools import get_recursive_attrs, order_lists
from optimeed.consolidate import analyse_sobol_convergence, SensitivityParameters, get_sensitivity_problem
from optimeed.visualize.process_mainloop import start_qt_mainloop
from SALib.analyze import sobol
import os
from optimeed.core import default_palette, Graphs, Data
from optimeed.visualize.graphs.pyqtgraph import TextItem
from optimeed.visualize.graphs.widget_graphsVisual import Widget_graphsVisual
from optimeed.visualize.mainWindow import MainWindow
from PyQt5.Qt import QFont
import numpy as np
import math


def analyse_sobol_plot_convergence(theDict, sobol='S1', title="", hold=True):
    """Plot convergence of the sobol indices.

    :param theDict: Dictionary containing sobol indices
    :param sobol: Key of the dictionary to investigate
    :param title: Title of the convergence window
    :param hold: If true, this function will be blocking (otherwise use start_qt_mainloop)
    :return: window containing convergence graphs
    """
    theGraphs = Graphs()
    font = QFont("Arial", 10)

    palette = default_palette(len(theDict))

    g1 = theGraphs.add_graph()
    myWidgetGraphsVisuals = Widget_graphsVisual(theGraphs, highlight_last=False, refresh_time=-1, is_light=True)
    for index, key in enumerate(theDict):
        color = palette[index]

        x = theDict[key]['step']
        y1 = theDict[key][sobol]
        theGraphs.add_trace(g1, Data(x, y1, symbol=None, x_label="Sample size", y_label="Sobol indices ({})".format(sobol), color=color, xlim=[0, x[-1]*1.2]))
        myText = TextItem(theDict[key]['name'], color=color, anchor=(0, 0.5))
        myText.setPos(x[-1], y1[-1])
        myText.setFont(font)
        myText.setParentItem(myWidgetGraphsVisuals.get_graph(g1).theWGPlot.vb)
        myWidgetGraphsVisuals.get_graph(g1).add_feature(myText)
        myWidgetGraphsVisuals.get_graph(g1).set_title(title)
    myWindow = MainWindow([myWidgetGraphsVisuals])  # A Window (that will contain the widget)
    myWindow.run(hold)
    return myWindow


def analyse_sobol_plot_indices(theSensitivityParameters: SensitivityParameters, objectives, title='', hold=True):
    """¨Plot first and total order sobol indices.

    :param theSensitivityParameters: Parameters used for sensitivity study
    :param objectives: List of evaluated objectives to analyse
    :param title: Title of the window
    :param hold: If true, this function will be blocking (otherwise use plt.show())
    :return:
    """
    problem_SALib = get_sensitivity_problem(theSensitivityParameters.get_optivariables())
    nb_params = len(theSensitivityParameters.get_optivariables())

    max_N = math.floor(len(objectives) / (2*nb_params + 2))
    Si = sobol.analyze(problem_SALib, np.array(objectives[0:max_N*(2*nb_params+2)]))

    _, ordered_S1 = order_lists(Si['S1'], list(range(nb_params)))
    _, ordered_ST = order_lists(Si['ST'], list(range(nb_params)))
    ordered_S1.reverse()
    ordered_ST.reverse()

    order = ordered_ST

    # width of the bars
    barWidth = 0.3

    bars1 = [Si['S1'][map_index] for map_index in order]
    bars2 = [Si['ST'][map_index] for map_index in order]
    yer1 = [Si['S1_conf'][map_index] for map_index in order]
    yer2 = [Si['ST_conf'][map_index] for map_index in order]

    labels = [theSensitivityParameters.get_optivariables()[map_index].get_attribute_name() for map_index in order]
    indices = list(range(len(bars1)))
    # labels = indices
    r1 = [x - barWidth/2 for x in indices]
    r2 = [x + barWidth/2 for x in indices]

    def split_labels(label_str, level=2):
        splitted = label_str.split('.')
        try:
            return '.'.join(splitted[-level:])
        except IndexError:
            return label_str
    plt.figure()
    plt.bar(r1, bars1, width=barWidth, color='blue', edgecolor='black', yerr=yer1, capsize=7, label='First index (S1)')
    plt.bar(r2, bars2, width=barWidth, color='cyan', edgecolor='black', yerr=yer2, capsize=7, label='Total index (ST)')
    plt.xticks(indices, map(split_labels, labels), rotation=30, ha="right")
    plt.ylabel('Sobol Index')
    plt.legend()
    plt.grid()
    plt.title(title)
    plt.show(block=hold)


def analyse_sobol_plot_2ndOrder_indices(theSensitivityParameters: SensitivityParameters, objectives, title='', hold=True):
    """¨Plot second order sobol indices. Args and kwargs are the same as analyse_sobol_plot_indices"""
    problem_SALib = get_sensitivity_problem(theSensitivityParameters.get_optivariables())
    nb_params = len(theSensitivityParameters.get_optivariables())

    names = [var.get_attribute_name() for var in theSensitivityParameters.get_optivariables()]

    max_N = math.floor(len(objectives) / (2 * nb_params + 2))
    Si = sobol.analyze(problem_SALib, np.array(objectives[0:max_N * (2 * nb_params + 2)]))

    matrix = Si["S2"]
    for i in range(nb_params):
        for j in range(i, nb_params):
            matrix[j, i] = matrix[i, j]

    fig, ax = plt.subplots()
    im = ax.imshow(matrix, cmap="Blues")
    cbar = ax.figure.colorbar(im, ax=ax)

    # Show all ticks and label them with the respective list entries
    slope = (4-12)/(75-25)
    offset = 12-slope*25
    max_len = max([len(word) for word in names])
    fontsize = max(6, int(max_len*slope + offset))
    fontsize = min(12, fontsize)

    ax.set_xticks(np.arange(nb_params), labels=names)
    ax.set_yticks(np.arange(nb_params), labels=names)
    plt.tick_params(axis='both', which='major', labelsize=fontsize)

    # # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=70, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    for i in range(len(names)):
        for j in range(len(names)):
            max_value = np.nanmax(matrix)
            curr_value_relative = matrix[i, j]/max_value
            color = "white" if curr_value_relative > 0.7 else "black"
            text = ax.text(j, i, "{:.3f}".format(matrix[i, j]),
                           ha="center", va="center", color=color)

    ax.set_title(title)
    fig.tight_layout()
    plt.show(block=hold)


class SensitivityDisplayer(QtWidgets.QMainWindow):
    """GUI to display a sensitivity analysis."""
    def __init__(self):
        super().__init__()  #
        uic.loadUi(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'displaySensitivity_gui.ui'), self)

        self.studies = list()
        self.curr_study = None

        self.combo_collection.currentIndexChanged.connect(self._set_study)
        # Todo
        self.button_get_S1_conv.clicked.connect(self._get_S1_conv)
        self.button_get_ST_conv.clicked.connect(self._get_ST_conv)
        self.button_get_indices.clicked.connect(self._get_sobol_indices)

        self.show()
        self.initialized = False
        self._windowsHolder = list()

    def add_study(self, theCollection, theParameters, name):
        """Add sensitivity study to the GUI

        :param theCollection: Results of the sensitivity study
        :param theParameters: Parameters of the sensitivity study
        :param name: Name (for the GUI) of the sensitivity study
        :return:
        """
        self.studies.append((theCollection, theParameters, name))
        self.combo_collection.addItem(name)
        if not self.initialized:
            self.combo_collection.setCurrentIndex(0)
            self.initialized = True

    def _set_study(self, index):
        self.curr_study = self.studies[index]
        first_data = self.curr_study[0].get_data_at_index(0)
        self.list_attributes.set_list(get_recursive_attrs(first_data))

    def _get_sobol_indices(self):
        collection, parameters, name = self.curr_study
        attribute_selected = self.list_attributes.get_name_selected()
        values_selected = collection.get_list_attributes(attribute_selected)
        analyse_sobol_plot_indices(parameters, values_selected, title="{} | {}".format(name, attribute_selected), hold=False)
        analyse_sobol_plot_2ndOrder_indices(parameters, values_selected, title="{} | {}".format(name, attribute_selected), hold=False)

    def _get_S1_conv(self):
        collection, parameters, name = self.curr_study
        attribute_selected = self.list_attributes.get_name_selected()
        win = analyse_sobol_plot_convergence(analyse_sobol_convergence(parameters, collection.get_list_attributes(attribute_selected), stepsize=10),
                                             sobol="S1", title="{} | {}".format(name, attribute_selected), hold=False)
        self._windowsHolder.append(win)

    def _get_ST_conv(self):
        collection, parameters, name = self.curr_study
        attribute_selected = self.list_attributes.get_name_selected()
        win = analyse_sobol_plot_convergence(analyse_sobol_convergence(parameters, collection.get_list_attributes(attribute_selected), stepsize=10),
                                             sobol="ST", title="{} | {}".format(name, attribute_selected), hold=False)
        self._windowsHolder.append(win)


    @staticmethod
    def run():
        start_qt_mainloop()




from .deviceDrawerInterface import DeviceDrawerInterface
from .materials import MaterialRenderingProperties, Emerald_material, Yellow_Emerald_material, Brass_material , \
    Bronze_material, Silver_material, Steel_material , Copper_material, \
    Chrome_material, Blue_material , Red_material , Green_material, Cyan_material, Pink_material


from .pyqtgraph import mkPen, mkBrush
from PyQt5 import QtCore, QtGui
from .colormap_pyqtgraph import matplotlib_colormap_to_pg_colormap


default_colormap = matplotlib_colormap_to_pg_colormap('jet')


def _normalize_colors(z):
    # Get colors to set
    min_value = min(z)
    max_value = max(z)
    colors = [(0, 0, 0)] * len(z)
    for k, value in enumerate(z):
        normalized_value = (value - min_value) / (max_value - min_value)
        colors[k] = default_colormap.mapToQColor(normalized_value)
    return colors


class TraceVisual(QtCore.QObject):
    """Defines a trace in a graph."""
    signal_must_update = QtCore.pyqtSignal()

    class _ModifiedPaintElem:
        """Hidden class to manage brushes or pens"""
        def __init__(self):
            self.changed = dict()  # {index_plot, newPaintElem}

        def add_modified_paintElem(self, index, newPaintElem):
            self.changed[index] = newPaintElem

        def modify_paintElems(self, paintElemsIn_List):
            """Apply transformation to paintElemsIn_List.

            :param: paintElemsIn_List: list of brushes or pens to modify
            :return: False if nothing has been modified, True is something has been modified
            """
            if not len(self.changed):
                return False

            for index in self.changed:
                try:
                    paintElemsIn_List[index] = self.changed[index]
                except IndexError:
                    pass
            return True

        def reset_paintElem(self, index):
            """ Remove transformation of point index"""
            self.changed.pop(index, None)

        def reset(self):
            self.__init__()

    def __init__(self, theData, theWGPlot, highlight_last):
        """
        :param theData: from class Data. Contains all the informations to plot (x, y, z, transforms, etc.)
        :param theWGPlot: holder of the trace. This is a plotWidget
        :param highlight_last: Boolean. If set to true, the last point in the data will be highlighted.
        """
        super().__init__()

        # set brush
        self.theBrushesModifier = self._ModifiedPaintElem()

        # set symbolPen
        self.theSymbolPensModifier = self._ModifiedPaintElem()

        # set symbols
        self.theSymbolModifier = self._ModifiedPaintElem()

        # set data
        self.theData = theData

        # set trace item (plotItem)
        if theData.get_legend():
            self.thePlotItem = theWGPlot.plot([float('inf')], name=theData.get_legend())
        else:
            self.thePlotItem = theWGPlot.plot([float('inf')])

        # set highlight last
        self.highlight_last = highlight_last
        self.drawPoints = True

        # signals
        self.signal_must_update.connect(self.updateTrace)

        # Performance issues:
        self._base_symbol_pen = None
        self._base_symbol_brush = None
        self._base_pen = None

        self.visible = True

    def hide_points(self):
        """Hide all the points"""
        self.drawPoints = False
        self.signal_must_update.emit()

    def get_color(self):
        """Get colour of the trace, return tuple (r,g,b)"""
        return self.get_data().get_color()

    def set_color(self, color):
        """Set colour of the trace, argument as tuple (r,g,b)"""
        self.get_data().set_color(color)
        self.thePlotItem.scatter.opts['brush'] = color

    def get_base_symbol_brush(self):
        """Get symbol brush configured for this trace, return `pg.QBrush`"""
        if self._base_symbol_brush is None:
            self._base_symbol_brush = mkBrush(self.get_data().get_color_alpha())
        return self._base_symbol_brush

    def get_base_pen(self):
        """Get pen configured for this trace, return `pg.QPen`"""
        if self._base_pen is None:
            pen = mkPen(self.get_color(), width=self.get_data().get_width())
            if self.theData.isScattered():
                pen.setStyle(QtCore.Qt.NoPen)
            else:
                self.set_pen_linestyle(pen, self.theData.get_linestyle())
            self._base_pen = pen
        return self._base_pen

    def get_base_symbol_pen(self):
        """Get symbol pen configured for this trace, return`pg.QPen`"""
        if self._base_symbol_pen is None:
            if self.get_data().symbol_isfilled():
                if self.get_data().get_symbolOutline() == 1:
                    return None
                init_color = QtGui.QColor(*self.get_color())
                darker_color = init_color.darker(int(self.get_data().get_symbolOutline()*100))
                alpha = self.get_data().get_alpha()
                if alpha < 255:
                    darker_color.setAlpha(self.get_data().get_alpha())
                thePen = mkPen(darker_color, isCosmetic=True)
                # thePen = mkPen(QtGui.QColor(*self.get_color()).darker(int(self.get_data().get_symbolOutline()*100)), isCosmetic=True)  # 120 or 250 ? :)
            else:
                thePen = mkPen(self.get_color(), isCosmetic=False, width=self.get_data().get_width())
            self._base_symbol_pen = thePen
        return self._base_symbol_pen

    def get_base_symbol(self):
        """Get base symbol configured for this trace, return str of the symbol (e.g. 'o')"""
        return self.get_data().get_symbol()

    def get_symbol(self, size):
        """Get actual symbols for the trace. If the symbols have been modified: return a list which maps each points to a symbol.
        Otherwise: return :meth:TraceVisual.get_base_symbol()"""
        symbol = self.get_base_symbol()
        theSymbolList = [symbol] * size
        hasBeenModified = self.theSymbolModifier.modify_paintElems(theSymbolList)
        return symbol if not hasBeenModified else theSymbolList

    def updateTrace(self):
        """Forces the trace to refresh."""
        if self.visible:
            try:
                theData = self.get_data()
                x, y = theData.get_plot_data()
                if x and y:
                    z = theData.get_plot_meta(x, y)

                    if z is not None:
                        self.set_brushes(list(range(len(z))), list(map(QtGui.QBrush, _normalize_colors(z))), update=False)

                    if self.drawPoints:
                        theBrushes = self.get_brushes(len(x))
                        theSymbolPens = self.get_symbolPens(len(x))
                    else:
                        theBrushes = None
                        theSymbolPens = None

                    theBasePen = self.get_base_pen()
                    self.thePlotItem.setData(x, y, symbolBrush=theBrushes, symbol=self.get_symbol(len(x)), symbolPen=theSymbolPens, symbolSize=self.get_data().get_symbolsize(), pen=theBasePen)  # symbolPen = None for no outline
                else:
                    self.thePlotItem.setData([], [], symbolBrush=None)

            except ValueError:
                pass
        else:
            self.thePlotItem.setData([], [], symbolBrush=None)

    def get_length(self):
        """Return number of data to plot"""
        x = self.get_data().get_plot_data()
        return len(x)

    def hide(self):
        """Hides the trace"""
        self.visible = False
        self.thePlotItem.clear()
        self.signal_must_update.emit()

    def show(self):
        """Shows the trace"""
        self.visible = True
        self.signal_must_update.emit()

    def toggle(self, boolean):
        """Toggle the trace (hide/show)"""
        if boolean:
            self.show()
        else:
            self.hide()

    def get_data(self):
        """Get data to plot :class:`~optimeed.visualize.graphs.Graphs.Data`"""
        return self.theData

    def get_brushes(self, size):
        """Get actual brushes for the trace (=symbol filling). return a list which maps each points to a symbol brush"""
        symbolBrush = self.get_base_symbol_brush() if self.get_data().symbol_isfilled() else None
        theBrushList = [symbolBrush] * size

        if self.highlight_last:
            theBrushList[-1] = mkBrush('y')
        hasBeenModified = self.theBrushesModifier.modify_paintElems(theBrushList)
        if not hasBeenModified and not self.highlight_last:
            return symbolBrush
        return theBrushList

    def set_brush(self, indexPoint, newbrush, update=True):
        """Set the symbol brush for a specific point:

        :param indexPoint: Index of the point (in the graph) to modify
        :param newbrush: either QBrush or tuple (r, g, b) of the new brush
        :param update: if True, update the trace afterwards. This is slow operation."""
        if isinstance(newbrush, tuple):
            newbrush = mkBrush(newbrush)
        self.theBrushesModifier.add_modified_paintElem(indexPoint, newbrush)
        if update:
            self.signal_must_update.emit()

    def set_symbol(self, indexPoint, newSymbol, update=True):
        """Set the symbol shape for a specific point:

        :param indexPoint: Index of the point (in the graph) to modify
        :param newSymbol: string of the new symbol (e.g.: 'o')
        :param update: if True, update the trace afterwards. This is slow operation."""
        self.theSymbolModifier.add_modified_paintElem(indexPoint, newSymbol)
        if update:
            self.signal_must_update.emit()

    def set_brushes(self, list_indexPoint, list_newbrush, update=True):
        """Same as :meth:`~TraceVisual.set_brush` but by taking a list as input"""
        if not isinstance(list_newbrush, list):
            list_newbrush = [list_newbrush] * len(list_indexPoint)
        for i in range(len(list_indexPoint)):
            self.set_brush(list_indexPoint[i], list_newbrush[i], update=False)
        if update:
            self.signal_must_update.emit()

    def reset_brush(self, indexPoint, update=True):
        """Reset the brush of the point indexpoint"""
        self.theBrushesModifier.reset_paintElem(indexPoint)
        if update:
            self.signal_must_update.emit()

    def reset_brushes(self, list_indexPoint, update=True):
        """Same as :meth:`~TraceVisual.reset_brush` but by taking a list as input"""
        [self.reset_brush(idPoint, update=False) for idPoint in list_indexPoint]
        if update:
            self.signal_must_update.emit()
            
    def reset_all_brushes(self, update=True):
        """Reset all the brushes"""
        self.theBrushesModifier.reset()
        if update:
            self.signal_must_update.emit()

    def reset_symbol(self, indexPoint, update=True):
        """Reset the symbol shape of the point indexpoint"""
        self.theSymbolModifier.reset_paintElem(indexPoint)
        if update:
            self.signal_must_update.emit()

    def get_symbolPens(self, size):
        """Get actual symbol pens for the trace (=symbol outline). return a list which maps each points to a symbol pen"""
        thePenList = [self.get_base_symbol_pen()] * size
        hasBeenModified = self.theSymbolPensModifier.modify_paintElems(thePenList)
        if not hasBeenModified:
            return self.get_base_symbol_pen()
        return thePenList

    def set_symbolPen(self, indexPoint, newPen, update=True):
        """Set the symbol shape for a specific point:

        :param indexPoint: Index of the point (in the graph) to modify
        :param newPen: QPen item or tuple of the color (r,g,b)
        :param update: if True, update the trace afterwards. This is slow operation."""
        if isinstance(newPen, tuple):
            newPen = mkPen(newPen)
        self.theSymbolPensModifier.add_modified_paintElem(indexPoint, newPen)
        if update:
            self.signal_must_update.emit()

    def set_symbolPens(self, list_indexPoint, list_newpens, update=True):
        """Same as :meth:`~TraceVisual.set_symbolPen` but by taking a list as input"""
        if not isinstance(list_newpens, list):
            list_newpens = [list_newpens] * len(list_indexPoint)
        for i in range(len(list_indexPoint)):
            self.set_symbolPen(list_indexPoint[i], list_newpens[i], update=False)
        if update:
            self.signal_must_update.emit()

    def reset_symbolPen(self, indexPoint, update=True):
        """Reset the symbol pen of the point indexpoint"""
        self.theSymbolPensModifier.reset_paintElem(indexPoint)
        if update:
            self.signal_must_update.emit()

    def reset_symbolPens(self, list_indexPoint, update=True):
        """Same as :meth:`~TraceVisual.reset_symbolPen` but by taking a list as input"""
        [self.reset_symbolPen(idPoint, update=False) for idPoint in list_indexPoint]
        if update:
            self.signal_must_update.emit()

    def reset_all_symbolPens(self, update=True):
        """Reset all the symbol pens"""
        self.theSymbolPensModifier.reset()
        if update:
            self.signal_must_update.emit()

    @staticmethod
    def set_pen_linestyle(thePen, linestyle):
        """Transform a pen for dashed lines:

        :param thePen: QPen item
        :param linestyle: str (e.g.: '.', '.-', '--', ...)
        """
        if linestyle in ['.', ':', '..']:
            thePen.setDashPattern([0.1, 2])
            thePen.setCapStyle(QtCore.Qt.RoundCap)
            thePen.setWidthF(thePen.width()*1.75)
        elif linestyle in ['.-', '-.']:
            thePen.setCapStyle(QtCore.Qt.RoundCap)
            thePen.setDashPattern([3, 2, 0.1, 2])
            thePen.setWidthF(thePen.width()*1.25)
        elif linestyle in ['-..', '.-.', '..-']:
            thePen.setDashPattern([0.1, 2, 3, 2, 0.1, 2])
            thePen.setWidthF(thePen.width()*1.25)
        elif linestyle in ['--']:
            thePen.setDashPattern([2.5, 2])
            thePen.setWidthF(thePen.width()*1.1)
        else:
            thePen.setStyle(QtCore.Qt.SolidLine)

    def get_point(self, indexPoint):
        """Return object pyqtgraph.SpotItem"""
        return self.thePlotItem.scatter.points()[indexPoint]



from .onclickInterface import OnclickInterface
from PyQt5.QtWidgets import QApplication


class Onclick_copySomething(OnclickInterface):
    """On Click: copy something"""

    def __init__(self, theDataLink, functionStrFromDevice):
        """
        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        :param functionStrFromDevice: simple method that takes a device (point that is clicked) as argument.
        """
        self.theLinkDataGraph = theDataLink
        self.function = functionStrFromDevice

    def graph_clicked(self, the_graph_visual, index_graph, index_trace, indices_points):
        theDevice = self.theLinkDataGraph.get_clicked_item(index_graph, index_trace, indices_points[0])

        # Add the point to the collection exporter
        theStr = self.function(theDevice)
        cb = QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(theStr, mode=cb.Clipboard)

    def get_name(self):
        return "Copy in clipboard Str"
